/**
 * LibreSource
 * Copyright (C) 2004-2008 Artenum SARL / INRIA
 * http://www.libresource.org - contact@artenum.com
 *
 * This file is part of the LibreSource software, 
 * which can be used and distributed under license conditions.
 * The license conditions are provided in the LICENSE.TXT file 
 * at the root path of the packaging that enclose this file. 
 * More information can be found at 
 * - http://dev.libresource.org/home/license
 *
 * Initial authors :
 *
 * Guillaume Bort / INRIA
 * Francois Charoy / Universite Nancy 2
 * Julien Forest / Artenum
 * Claude Godart / Universite Henry Poincare
 * Florent Jouille / INRIA
 * Sebastien Jourdain / INRIA / Artenum
 * Yves Lerumeur / Artenum
 * Pascal Molli / Universite Henry Poincare
 * Gerald Oster / INRIA
 * Mariarosa Penzi / Artenum
 * Gerard Sookahet / Artenum
 * Raphael Tani / INRIA
 *
 * Contributors :
 *
 * Stephane Bagnier / Artenum
 * Amadou Dia / Artenum-IUP Blois
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 */
package org.libresource.kernel.ejb;

import org.jivesoftware.smack.XMPPConnection;

import org.libresource.Libresource;
import org.libresource.LibresourceEvent;
import org.libresource.LibresourceException;
import org.libresource.LibresourceResourceIdentifier;
import org.libresource.LibresourceResourceValue;
import org.libresource.LibresourceService;
import org.libresource.LibresourceServiceBase;

import org.libresource.kernel.KernelConstants;
import org.libresource.kernel.KernelServiceException;
import org.libresource.kernel.LibresourceSecurityException;
import org.libresource.kernel.URINotExistException;
import org.libresource.kernel.ejb.model.AclValue;
import org.libresource.kernel.ejb.model.SymbolicLinkResourceValue;
import org.libresource.kernel.interfaces.NodeLocal;
import org.libresource.kernel.interfaces.NodeLocalHome;
import org.libresource.kernel.interfaces.SymbolicLinkResourceLocal;
import org.libresource.kernel.mappers.PrincipalMapper;
import org.libresource.kernel.util.NodeUtil;
import org.libresource.kernel.util.SymbolicLinkResourceUtil;

import org.libresource.membership.MembershipConstants;
import org.libresource.membership.ejb.model.ProfileResourceValue;
import org.libresource.membership.interfaces.MembershipService;
import org.libresource.membership.interfaces.ProfileResourceLocal;
import org.libresource.membership.util.ProfileResourceUtil;

import org.libresource.scheduler.LibresourceJob;
import org.libresource.scheduler.LibresourceScheduler;

import org.libresource.search.LibresourceSearch;
import org.libresource.search.LibresourceSearchResult;

import org.quartz.CronTrigger;
import org.quartz.JobDetail;
import org.quartz.Scheduler;
import org.quartz.SimpleTrigger;
import org.quartz.Trigger;

import java.io.File;

import java.net.URI;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Stack;
import java.util.StringTokenizer;
import java.util.Vector;

import javax.ejb.ObjectNotFoundException;

import javax.mail.Session;
import javax.mail.Transport;
import javax.mail.internet.InternetAddress;
import javax.mail.internet.MimeMessage;

import javax.naming.Context;
import javax.naming.InitialContext;

import javax.rmi.PortableRemoteObject;

/**
 * LibreSource
 * 
 * @author <a href="mailto:bort@loria.fr">Guillaume Bort </a>- <a
 *         href="http://www.inria.fr">INRIA Lorraine </a>
 * 
 * The LibreSource Kernel Service
 * 
 * @libresource.service name="Kernel"
 * 
 * @ejb.env-entry name="uri-scheme" value="ls"
 * @ejb.env-entry name="principal-mapper"
 *                value="org.libresource.kernel.mappers.MembershipPrincipalMapper"
 * @ejb.env-entry name="principal-mapper-options" value=""
 * @ejb.env-entry name="server-config-uri" value="/ServerConfig"
 * @ejb.env-entry name="libresource-version" value="2"
 */
public abstract class KernelServiceBean extends LibresourceServiceBase {
	private static Object ANONYMOUS = new Object();

	private transient String uriScheme;

	private transient String uriHost;

	private transient String serverUrl;

	private transient String authUrl;

	private transient String libresourceVersion;

	private transient PrincipalMapper principalMapper;

	private transient Hashtable principals = new Hashtable();

	private transient String indexationDataDir;

	private transient URI serverConfigURI;

	private transient Session mailSession;

	public KernelServiceBean() {
		try {
			LibresourceScheduler.getInstance();
		} catch (Exception e) {
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public String getIndexationDataDir() throws KernelServiceException, LibresourceSecurityException {
		try {
			if (indexationDataDir == null) {
				// indexationDataDir = (String) new
				// InitialContext().lookup("java:comp/env/indexation-data-dir");
				indexationDataDir = org.libresource.Libresource.getLibresourceConfiguration("kernel.indexationDataDir");

				if ((indexationDataDir == null) || indexationDataDir.trim().equals("")) {
					File tempDirectory = new File(System.getProperty("java.io.tmpdir"), "libresource-index");
					tempDirectory.mkdirs();
					indexationDataDir = tempDirectory.getAbsolutePath();
				}
			}

			return indexationDataDir;
		} catch (Exception e) {
			throw new KernelServiceException("Error in getIndexationDataDir : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public URI getServerConfigURI() throws KernelServiceException, LibresourceSecurityException {
		try {
			if (serverConfigURI == null) {
				serverConfigURI = normalizeURI(new URI((String) new InitialContext().lookup("java:comp/env/server-config-uri")));
			}

			return serverConfigURI;
		} catch (Exception e) {
			throw new KernelServiceException("Error in getServerConfigURI : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Required"
	 */
	public void scheduleJob(LibresourceJob job, String description, Date date) throws KernelServiceException {
		try {
			// get scheduler
			Scheduler scheduler = LibresourceScheduler.getInstance().getScheduler();

			// remove old job with same name ...
			scheduler.deleteJob(job.getName(), Scheduler.DEFAULT_GROUP);

			// create new job
			JobDetail jobDetail = LibresourceScheduler.createJobDetail(job, description);

			// create trigger
			SimpleTrigger trigger = new SimpleTrigger("trigger." + job.getName(), Scheduler.DEFAULT_GROUP, date, null, 0, 0L);

			// schedule job
			scheduler.scheduleJob(jobDetail, trigger);
		} catch (Exception e) {
			ctx.setRollbackOnly();
			throw new KernelServiceException("Error in scheduleJob " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Required"
	 */
	public void scheduleJob(LibresourceJob job, String description, String cronExpression) throws KernelServiceException {
		try {
			// get scheduler
			Scheduler scheduler = LibresourceScheduler.getInstance().getScheduler();

			// remove old job with same name ...
			scheduler.deleteJob(job.getName(), Scheduler.DEFAULT_GROUP);

			// create new job
			JobDetail jobDetail = LibresourceScheduler.createJobDetail(job, description);

			// create trigger
			CronTrigger trigger = new CronTrigger("trigger." + job.getName(), Scheduler.DEFAULT_GROUP, cronExpression);
			trigger.setMisfireInstruction(CronTrigger.MISFIRE_INSTRUCTION_DO_NOTHING);

			// schedule job
			scheduler.scheduleJob(jobDetail, trigger);
		} catch (Exception e) {
			ctx.setRollbackOnly();
			throw new KernelServiceException("Error in scheduleJob " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Required"
	 */
	public void unScheduleJob(String jobName) throws KernelServiceException {
		try {
			// get scheduler
			Scheduler scheduler = LibresourceScheduler.getInstance().getScheduler();

			// remove old job with same name ...
			scheduler.deleteJob(jobName, Scheduler.DEFAULT_GROUP);
		} catch (Exception e) {
			ctx.setRollbackOnly();
			throw new KernelServiceException("Error in unScheduleJob " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public String[] listScheduledJobs() throws KernelServiceException {
		try {
			// get scheduler
			Scheduler scheduler = LibresourceScheduler.getInstance().getScheduler();

			// remove old job with same name ...
			return scheduler.getJobNames(Scheduler.DEFAULT_GROUP);
		} catch (Exception e) {
			throw new KernelServiceException("Error in listScheduledJobs " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public String getJobDescription(String jobName) throws KernelServiceException {
		try {
			// get scheduler
			Scheduler scheduler = LibresourceScheduler.getInstance().getScheduler();

			// remove old job with same name ...
			return scheduler.getJobDetail(jobName, Scheduler.DEFAULT_GROUP).getDescription();
		} catch (Exception e) {
			throw new KernelServiceException("Error in getJobDescription " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public String getTriggersInformation(String jobName) throws KernelServiceException {
		try {
			// get scheduler
			Scheduler scheduler = LibresourceScheduler.getInstance().getScheduler();

			// remove old job with same name ...
			Trigger[] triggers = scheduler.getTriggersOfJob(jobName, Scheduler.DEFAULT_GROUP);

			if (triggers.length > 0) {
				if (triggers[0] instanceof SimpleTrigger) {
					SimpleTrigger theTrigger = (SimpleTrigger) triggers[0];

					return "Fire at " + theTrigger.getNextFireTime();
				} else if (triggers[0] instanceof CronTrigger) {
					CronTrigger theTrigger = (CronTrigger) triggers[0];

					return "Cron :  " + theTrigger.getCronExpression() + " (next fire at " + theTrigger.getNextFireTime() + ")";
				}
			}

			return "No informations";
		} catch (Exception e) {
			throw new KernelServiceException("Error in getTriggersInformation " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public String getUriScheme() throws KernelServiceException {
		try {
			if (uriScheme == null) {
				uriScheme = (String) new InitialContext().lookup("java:comp/env/uri-scheme");
			}

			return uriScheme;
		} catch (Exception e) {
			throw new KernelServiceException("Error in getUriScheme " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public String getServerUrl() throws KernelServiceException {
		try {
			if (serverUrl == null) {
				// serverUrl = (String) new
				// InitialContext().lookup("java:comp/env/server-url");
				serverUrl = org.libresource.Libresource.getLibresourceConfiguration("kernel.serverUrl");
			}

			return serverUrl;
		} catch (Exception e) {
			throw new KernelServiceException("Error in getServerUrl " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public String getUriHost() throws KernelServiceException {
		try {
			if (uriHost == null) {
				uriHost = new URI(getServerUrl()).getHost();

				int port = new URI(getServerUrl()).getPort();

				if (port != -1) {
					uriHost = uriHost + ":" + port;
				}
			}

			return uriHost;
		} catch (Exception e) {
			throw new KernelServiceException("Error in getUriHost " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public String getAuthUrl() throws KernelServiceException {
		try {
			if (authUrl == null) {
				// authUrl = (String) new
				// InitialContext().lookup("java:comp/env/auth-url");
				authUrl = org.libresource.Libresource.getLibresourceConfiguration("kernel.authUrl");
			}

			return authUrl;
		} catch (Exception e) {
			throw new KernelServiceException("Error in getAuthUrl " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public String getLibresourceVersion() throws KernelServiceException {
		try {
			if (libresourceVersion == null) {
				libresourceVersion = (String) new InitialContext().lookup("java:comp/env/libresource-version");
			}

			return libresourceVersion;
		} catch (Exception e) {
			throw new KernelServiceException("Error in getLibresourceVersion " + e.getMessage(), e);
		}
	}

	// mails

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public Session getMailSession() throws Exception {
		if (mailSession == null) {
			mailSession = (Session) PortableRemoteObject.narrow(new InitialContext().lookup(Libresource.getLibresourceConfiguration("mail.lookup")),
					Session.class);
		}

		return mailSession;
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public void sendEmail(String userId, String mail, String subject) throws Exception {
		// TODO check if profile exist
		MembershipService membershipService = (MembershipService) Libresource.getService(MembershipConstants.SERVICE);
		ProfileResourceValue user = membershipService.systemGetProfile(new URI(membershipService.getUsersRootURI() + "/" + userId));
		MimeMessage m = new MimeMessage(getMailSession());
		m.setFrom();
		m.setRecipients(javax.mail.Message.RecipientType.TO, user.getEmail());
		m.setSubject(subject);
		m.setSentDate(new java.util.Date());
		m.setContent("For user " + user.getFullName() + "\r\n\r\n" + mail, "text/plain");
		Transport.send(m);
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public void sendEmail(String from, String userId, String mail, String subject) throws Exception {
		// TODO check if profile exist
		MembershipService membershipService = (MembershipService) Libresource.getService(MembershipConstants.SERVICE);
		ProfileResourceValue user = membershipService.systemGetProfile(new URI(membershipService.getUsersRootURI() + "/" + userId));
		MimeMessage m = new MimeMessage(getMailSession());
		m.setFrom(new InternetAddress(from));
		m.setRecipients(javax.mail.Message.RecipientType.TO, user.getEmail());
		m.setSubject(subject);
		m.setSentDate(new java.util.Date());
		m.setContent("For user " + user.getFullName() + "\r\n\r\n" + mail, "text/plain");
		Transport.send(m);
	}

	// nodes management

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Required"
	 */
	public void bind(LibresourceResourceIdentifier resourceIdentifier, URI uri, String resourceName) throws KernelServiceException,
			LibresourceSecurityException, URINotExistException {
		try {
			createURI(uri);

			if (!checkIfCanChangeRights(uri)) {
				throw new LibresourceSecurityException(uri, "OWNER or CRUD");
			}

			String path = normalizeAbsoluteURIPath(uri);

			try {
				try {
					NodeUtil.getLocalHome().findByBindedResource(resourceIdentifier.serialize());
					throw new Exception("This resource is already binded with another URI.");
				} catch (ObjectNotFoundException e) {
					//
				}

				NodeLocal pathNodeLocal = NodeUtil.getLocalHome().getByPath(path);

				if (pathNodeLocal.getBindedResourceIdentifier() != null) {
					throw new Exception("A resource is already binded with this URI " + uri);
				}

				pathNodeLocal.setBindedResourceIdentifier(resourceIdentifier.serialize());

				// filter ACLs
				HashMap acls = pathNodeLocal.getAcls();

				for (Iterator i = acls.keySet().iterator(); i.hasNext();) {
					URI key = (URI) i.next();
					String[] permissions = (String[]) acls.get(key);
					acls.put(key, filterPermissions(resourceIdentifier, permissions));
				}

				pathNodeLocal.setAcls(acls);
				pathNodeLocal.setShortName(resourceName);
				pathNodeLocal.setUpdateDate(new Date());
			} catch (ObjectNotFoundException e) {
				throw new URINotExistException(normalizeURI(uri));
			}

			// indexation
			Libresource.index(uri);

			// event
			Libresource.throwEvent(new LibresourceEvent(uri, resourceIdentifier, getConnectedResource(), KernelConstants.EVENT_BIND));
		} catch (LibresourceSecurityException se) {
			ctx.setRollbackOnly();
			throw se;
		} catch (URINotExistException se) {
			ctx.setRollbackOnly();
			throw se;
		} catch (Exception e) {
			ctx.setRollbackOnly();
			throw new KernelServiceException("Error in bind : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Required"
	 */
	public void move(URI from, URI to) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			if (exist(to)) {
				throw new Exception("URI " + to + " already exists");
			}

			if (isRoot(from)) {
				throw new Exception("Can't move the root node");
			}

			String fromPath = normalizeAbsoluteURIPath(from);
			String toPath = normalizeAbsoluteURIPath(to);

			if (toPath.startsWith(fromPath + "/")) {
				throw new Exception("Cannot move URI " + from + " to a child node of itself");
			}

			String toPathParent = toPath.substring(0, toPath.lastIndexOf("/"));
			String toPathLastNode = toPath.substring(toPath.lastIndexOf("/") + 1, toPath.length());

			if (toPathParent.equals("")) {
				toPathParent = "/";
			}

			NodeLocal fromNode;
			NodeLocal toParentNode;

			try {
				fromNode = NodeUtil.getLocalHome().getByPath(fromPath);
			} catch (ObjectNotFoundException e) {
				throw new URINotExistException(normalizeURI(from));
			}

			String oldPathPart = fromNode.getPathPart();

			if (!checkSecurity(new URI(toPathParent), KernelConstants.SECURITY_CREATE)) {
				throw new LibresourceSecurityException(new URI(toPathParent), KernelConstants.SECURITY_CREATE);
			}

			if (!checkSecurity(from, KernelConstants.SECURITY_DELETE)) {
				throw new LibresourceSecurityException(from, KernelConstants.SECURITY_DELETE);
			}

			createURI(new URI(toPathParent));
			toParentNode = NodeUtil.getLocalHome().getByPath(toPathParent);

			NodeLocal parent = NodeUtil.getLocalHome().findByPrimaryKey(fromNode.getParent());
			fromNode.setParent(toParentNode.getId());
			fromNode.setPathPart(toPathLastNode);
			toParentNode.addChild(fromNode);
			parent.removeChild(oldPathPart);
			fromNode.updatePath();

			// event
			Libresource.throwEvent(new LibresourceEvent(from, KernelConstants.SERVICE, "Node", getConnectedResource(), KernelConstants.EVENT_MOVE_FROM));
			Libresource.throwEvent(new LibresourceEvent(to, KernelConstants.SERVICE, "Node", getConnectedResource(), KernelConstants.EVENT_MOVE_TO));
		} catch (LibresourceSecurityException se) {
			ctx.setRollbackOnly();
			throw se;
		} catch (URINotExistException se) {
			ctx.setRollbackOnly();
			throw se;
		} catch (Exception e) {
			ctx.setRollbackOnly();
			throw new KernelServiceException("Error in move : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public boolean exist(URI uri) throws KernelServiceException, LibresourceSecurityException {
		try {
			String path = normalizeAbsoluteURIPath(uri);

			try {
				NodeUtil.getLocalHome().getByPath(path);
			} catch (ObjectNotFoundException e) {
				return false;
			}

			return true;
		} catch (Exception e) {
			throw new KernelServiceException("Error in exist " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Required"
	 */
	public void unbind(URI uri) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			String path = normalizeAbsoluteURIPath(uri);

			try {
				NodeLocal pathNodeLocal = NodeUtil.getLocalHome().getByPath(path);

				if (!checkIfCanChangeRights(uri)) {
					throw new LibresourceSecurityException(uri, "OWNER or CRUD");
				}

				pathNodeLocal.setBindedResourceIdentifier(null);
				pathNodeLocal.setShortName("(empty node)");
				pathNodeLocal.setUpdateDate(new Date());
			} catch (ObjectNotFoundException e) {
				throw new URINotExistException(normalizeURI(uri));
			}

			// indexation
			Libresource.removeFromIndex(uri);

			// event
			Libresource.throwEvent(new LibresourceEvent(uri, KernelConstants.SERVICE, "Node", getConnectedResource(), KernelConstants.EVENT_UNBIND));
		} catch (LibresourceSecurityException se) {
			ctx.setRollbackOnly();
			throw se;
		} catch (URINotExistException se) {
			ctx.setRollbackOnly();
			throw se;
		} catch (Exception e) {
			ctx.setRollbackOnly();
			throw new KernelServiceException("Error in unbind : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public LibresourceResourceIdentifier lookup(URI uri) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			String path = normalizeAbsoluteURIPath(uri);

			try {
				NodeLocal pathNodeLocal = NodeUtil.getLocalHome().getByPath(path);

				/*
				 * if (!checkSecurity(uri,
				 * KernelConstants.SECURITY_KERNEL_READ)) { throw new
				 * LibresourceSecurityException(uri,
				 * KernelConstants.SECURITY_KERNEL_READ); }
				 */
				if (pathNodeLocal.getBindedResourceIdentifier() == null) {
					return null;
				}

				return LibresourceResourceIdentifier.deserialize(pathNodeLocal.getBindedResourceIdentifier());
			} catch (ObjectNotFoundException e) {
				throw new URINotExistException(normalizeURI(uri) + " :: " + e.getMessage());
			}
		} catch (LibresourceSecurityException se) {
			throw se;
		} catch (URINotExistException se) {
			throw se;
		} catch (Exception e) {
			throw new KernelServiceException("Error in lookup : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public String getShortName(URI uri) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			String path = normalizeAbsoluteURIPath(uri);

			try {
				NodeLocal pathNodeLocal = NodeUtil.getLocalHome().getByPath(path);

				// if (!checkSecurity(uri, "Kernel:READ")) {
				// return path.substring(path.lastIndexOf("/") + 1,
				// path.length());
				// }
				return pathNodeLocal.getShortName();
			} catch (ObjectNotFoundException e) {
				throw new URINotExistException(normalizeURI(uri) + " :: " + e.getMessage());
			}
		} catch (LibresourceSecurityException se) {
			throw se;
		} catch (URINotExistException se) {
			throw se;
		} catch (Exception e) {
			throw new KernelServiceException("Error in lookup : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Required"
	 */
	public void setShortName(URI uri, String name) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			String path = normalizeAbsoluteURIPath(uri);

			try {
				NodeLocal pathNodeLocal = NodeUtil.getLocalHome().getByPath(path);

				// Shouldn't be needed
				/*
				 * if (!checkSecurity(uri,
				 * KernelConstants.SECURITY_KERNEL_UPDATE)) { throw new
				 * LibresourceSecurityException(uri,
				 * KernelConstants.SECURITY_KERNEL_UPDATE); }
				 */
				pathNodeLocal.setShortName(name);
			} catch (ObjectNotFoundException e) {
				throw new URINotExistException(normalizeURI(uri) + " :: " + e.getMessage());
			}
		} catch (LibresourceSecurityException se) {
			throw se;
		} catch (URINotExistException se) {
			throw se;
		} catch (Exception e) {
			throw new KernelServiceException("Error in setShortName : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public URI getOwner(URI uri) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			String path = normalizeAbsoluteURIPath(uri);

			try {
				NodeLocal pathNodeLocal = NodeUtil.getLocalHome().getByPath(path);

				if (!checkSecurity(uri, KernelConstants.SECURITY_READ)) {
					throw new LibresourceSecurityException(uri, KernelConstants.SECURITY_READ);
				}

				if (pathNodeLocal.getOwner().trim().length() == 0) {
					return null;
				}

				NodeLocal owner = NodeUtil.getLocalHome().findByPrimaryKey(pathNodeLocal.getOwner());

				if (owner == null) {
					return null;
				}

				return normalizeURI(owner.getURI());
			} catch (ObjectNotFoundException e) {
				throw new URINotExistException(normalizeURI(uri));
			}
		} catch (LibresourceSecurityException se) {
			throw se;
		} catch (URINotExistException se) {
			throw se;
		} catch (Exception e) {
			throw new KernelServiceException("Error in getOwner : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Required"
	 */
	public void deleteURI(URI uri) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			String path = normalizeAbsoluteURIPath(uri);

			if (isRoot(uri)) {
				throw new Exception("Can't delete the root node");
			}

			try {
				NodeLocal nodeLocal = NodeUtil.getLocalHome().getByPath(path);

				if (!checkSecurity(uri, KernelConstants.SECURITY_DELETE)) {
					throw new LibresourceSecurityException(uri, KernelConstants.SECURITY_DELETE);
				}

				// event
				Libresource.throwEvent(new LibresourceEvent(uri, KernelConstants.SERVICE, "Node", getConnectedResource(), KernelConstants.EVENT_DELETE));
				nodeLocal.remove();

				/*
				 * Collection children = pathNodeLocal.getChildren2(); for
				 * (Iterator i = children.iterator(); i.hasNext();) {
				 * deleteURI(((NodeLocal) i.next()).getURI()); }
				 * 
				 * NodeUtil.getLocalHome().getByPath(pathNodeLocal.getParent()).removeChild(pathNodeLocal.getId());
				 * nodeLocal.remove();
				 */
			} catch (ObjectNotFoundException e) {
				throw new URINotExistException(normalizeURI(uri));
			}
		} catch (LibresourceSecurityException se) {
			ctx.setRollbackOnly();
			throw se;
		} catch (URINotExistException se) {
			ctx.setRollbackOnly();
			throw se;
		} catch (Exception e) {
			ctx.setRollbackOnly();
			throw new KernelServiceException("Error in deleteURI : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Required"
	 */
	public void createURI(URI uri) throws KernelServiceException, LibresourceSecurityException {
		try {
			NodeLocalHome pathNodeLocalHome = NodeUtil.getLocalHome();
			StringTokenizer tokenizer = new StringTokenizer(normalizeAbsoluteURIPath(uri), "/");
			ArrayList pathParts = new ArrayList();

			while (tokenizer.hasMoreTokens()) {
				pathParts.add(tokenizer.nextToken());
			}

			NodeLocal tempParent = null;

			try {
				tempParent = pathNodeLocalHome.findByPrimaryKey("666");
			} catch (ObjectNotFoundException e) {
				throw new KernelServiceException("root node must exist");
			}

			for (Iterator i = pathParts.iterator(); i.hasNext();) {
				String pathPart = (String) i.next();

				// tempPath.append(pathPart);
				String childid = (String) tempParent.getChildren().get(pathPart);

				if (childid == null) {
					if (!checkSecurity(tempParent.getURI(), KernelConstants.SECURITY_CREATE)) {
						throw new LibresourceSecurityException(tempParent.getURI(), KernelConstants.SECURITY_CREATE);
					}

					NodeLocal newNode = pathNodeLocalHome.create(tempParent, pathPart);
					copyAcls(tempParent, newNode);
					tempParent.addChild(newNode);
					tempParent = newNode;

					// try {
					newNode.setOwner(NodeUtil.getLocalHome().getByPath(normalizeAbsoluteURIPath(getConnectedResource())).getId());

					// newNode.setOwner(getConnectedResource().getId());
					// } catch (ObjectNotFoundException onfe) {
					// throw new KernelServiceException("Can't set this owner
					// (not found) : " + getConnectedResource());
					// }
				} else {
					tempParent = pathNodeLocalHome.findByPrimaryKey(childid);
				}
			}

			// event
			Libresource.throwEvent(new LibresourceEvent(uri, KernelConstants.SERVICE, "Node", getConnectedResource(), KernelConstants.EVENT_CREATE));
		} catch (KernelServiceException e) {
			ctx.setRollbackOnly();
			throw e;
		} catch (LibresourceSecurityException e) {
			ctx.setRollbackOnly();
			throw e;
		} catch (Exception e) {
			ctx.setRollbackOnly();
			throw new KernelServiceException("Error in createURI : " + e.getMessage(), e);
		}
	}

	/*
	 * private NodeLocal createNode(NodeLocal parent, String path) throws
	 * KernelServiceException, LibresourceSecurityException,
	 * URINotExistException { try { return
	 * NodeUtil.getLocalHome().create(parent, path); } catch (Exception e) {
	 * throw new KernelServiceException("Error in getParent : " +
	 * e.getMessage(), e); } }
	 */

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public URI getParent(URI uri) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			try {
				NodeLocal pathNodeLocal = NodeUtil.getLocalHome().getByPath(normalizeAbsoluteURIPath(uri));

				if (pathNodeLocal.getParent().trim().length() == 0) {
					return null;
				}

				NodeLocal parent = NodeUtil.getLocalHome().findByPrimaryKey(pathNodeLocal.getParent());

				if (parent == null) {
					return null;
				}

				return normalizeURI(parent.getURI());
			} catch (ObjectNotFoundException e) {
				throw new URINotExistException(normalizeURI(uri));
			}
		} catch (URINotExistException se) {
			throw se;
		} catch (Exception e) {
			throw new KernelServiceException("Error in getParent : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public URI getMyURI(LibresourceResourceIdentifier resourceIdentifier) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			NodeLocal pathNodeLocal = NodeUtil.getLocalHome().findByBindedResource(resourceIdentifier.serialize());

			return normalizeURI(pathNodeLocal.getURI());
		} catch (ObjectNotFoundException e) {
			throw new KernelServiceException("This resource is not binded");
		} catch (URINotExistException se) {
			throw se;
		} catch (Exception e) {
			throw new KernelServiceException("Error in getMyURI : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public boolean isRoot(URI uri) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			return normalizeAbsoluteURIPath(uri).equals("/");
		} catch (URINotExistException se) {
			throw se;
		} catch (Exception e) {
			throw new KernelServiceException("Error in isRoot : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public URI normalizeURI(URI uri) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			return new URI(getUriScheme() + "://" + getUriHost() + normalizeAbsoluteURIPath(uri));
		} catch (URINotExistException se) {
			throw se;
		} catch (Exception e) {
			throw new KernelServiceException("Error in normalizeURI : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Required"
	 */
	public void chown(URI uri, URI owner, boolean recursive) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			if (!checkIfCanChangeRights(uri)) {
				throw new LibresourceSecurityException(uri, "OWNER or CRUD");
			}

			systemChown(uri, owner, recursive);
		} catch (LibresourceSecurityException se) {
			ctx.setRollbackOnly();
			throw se;
		} catch (URINotExistException se) {
			ctx.setRollbackOnly();
			throw se;
		} catch (Exception e) {
			ctx.setRollbackOnly();
			throw new KernelServiceException("Error in chown : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public boolean checkIfCanChangeRights(URI user, URI uri) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			return (checkIfOwner(uri) || (checkSecurity(uri, KernelConstants.SECURITY_CREATE) && checkSecurity(uri, KernelConstants.SECURITY_DELETE)
					&& checkSecurity(uri, KernelConstants.SECURITY_UPDATE) && checkSecurity(uri, KernelConstants.SECURITY_READ)));
		} catch (URINotExistException se) {
			return false;
		} catch (Exception e) {
			ctx.setRollbackOnly();
			throw new KernelServiceException("Error in checkIfCanChangeRights : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public boolean checkIfCanChangeRights(URI uri) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		return checkIfCanChangeRights(getConnectedResource(), uri);
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Required"
	 */
	public void createAcl(URI forUri, URI resource, String[] permissions) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			NodeLocal forUriNode;
			NodeLocal resourceNode;

			try {
				forUriNode = NodeUtil.getLocalHome().getByPath(normalizeAbsoluteURIPath(forUri));
			} catch (ObjectNotFoundException e) {
				throw new URINotExistException(normalizeURI(forUri));
			}

			try {
				resourceNode = NodeUtil.getLocalHome().getByPath(normalizeAbsoluteURIPath(resource));
			} catch (ObjectNotFoundException e) {
				throw new URINotExistException(normalizeURI(resource));
			}

			if (!checkIfCanChangeRights(forUri)) {
				throw new LibresourceSecurityException(forUri, "OWNER or CRUD");
			}

			HashMap acls = forUriNode.getAcls();

			if (acls == null) {
				acls = new HashMap();
			}

			// Keep only owner that have at least on right
			String[] resultingPermission = filterPermissions(LibresourceResourceIdentifier.deserialize(forUriNode.getBindedResourceIdentifier()), permissions);

			if (resultingPermission.length > 0) {
				acls.put(resourceNode.getURI(), resultingPermission);
			}

			forUriNode.setAcls(acls);
			forUriNode.setUpdateDate(new Date());

			// event
			Libresource.throwEvent(new LibresourceEvent(forUri, KernelConstants.SERVICE, "Node", getConnectedResource(), KernelConstants.EVENT_CREATE_ACLS));
		} catch (LibresourceSecurityException se) {
			ctx.setRollbackOnly();
			throw se;
		} catch (URINotExistException se) {
			ctx.setRollbackOnly();
			throw se;
		} catch (Exception e) {
			ctx.setRollbackOnly();
			throw new KernelServiceException("Error in createAcl : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Required"
	 */
	public void createAclRecursively(URI forUri, URI resource, String[] permissions) throws KernelServiceException, LibresourceSecurityException,
			URINotExistException {
		try {
			NodeLocal forUriNode;
			NodeLocal resourceNode;

			try {
				forUriNode = NodeUtil.getLocalHome().getByPath(normalizeAbsoluteURIPath(forUri));
			} catch (ObjectNotFoundException e) {
				throw new URINotExistException(normalizeURI(forUri));
			}

			try {
				resourceNode = NodeUtil.getLocalHome().getByPath(normalizeAbsoluteURIPath(resource));
			} catch (ObjectNotFoundException e) {
				throw new URINotExistException(normalizeURI(resource));
			}

			if (!checkIfCanChangeRights(forUri)) {
				throw new LibresourceSecurityException(forUri, "OWNER or CRUD");
			}

			// recursive
			Collection children = forUriNode.getChildren2();

			for (Iterator i = children.iterator(); i.hasNext();) {
				NodeLocal child = (NodeLocal) i.next();
				createAclRecursively(child.getURI(), resource, permissions);
			}

			HashMap acls = forUriNode.getAcls();

			if (acls == null) {
				acls = new HashMap();
			}

			String[] existingPermissions = (String[]) acls.get(resourceNode.getURI());

			// keep specific existing permissions
			if (existingPermissions != null) {
				Vector v = new Vector();

				for (int i = 0; i < permissions.length; i++) {
					v.add(permissions[i]);
				}

				for (int i = 0; i < existingPermissions.length; i++) {
					if (!existingPermissions[i].startsWith(KernelConstants.SERVICE)) {
						v.add(existingPermissions[i]);
					}
				}

				String[] newPermissions = new String[v.size()];

				for (int i = 0; i < newPermissions.length; i++) {
					newPermissions[i] = (String) v.get(i);
				}

				permissions = newPermissions;
			}

			acls
					.put(resourceNode.getURI(), filterPermissions(LibresourceResourceIdentifier.deserialize(forUriNode.getBindedResourceIdentifier()),
							permissions));
			forUriNode.setAcls(acls);
			forUriNode.setUpdateDate(new Date());

			// event
			Libresource.throwEvent(new LibresourceEvent(forUri, KernelConstants.SERVICE, "Node", getConnectedResource(), KernelConstants.EVENT_CREATE_ACLS,
					permissions + " for " + resource));
		} catch (LibresourceSecurityException se) {
			ctx.setRollbackOnly();
			throw se;
		} catch (URINotExistException se) {
			ctx.setRollbackOnly();
			throw se;
		} catch (Exception e) {
			ctx.setRollbackOnly();
			throw new KernelServiceException("Error in createAcl : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Required"
	 */
	public void deleteAcl(URI forUri, URI resource, boolean recursive) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			if (!checkIfCanChangeRights(forUri)) {
				throw new LibresourceSecurityException(forUri, "OWNER or CRUD");
			}

			NodeLocal forUriNode;

			// NodeLocal resourceNode;
			try {
				forUriNode = NodeUtil.getLocalHome().getByPath(normalizeAbsoluteURIPath(forUri));
			} catch (ObjectNotFoundException e) {
				throw new URINotExistException(normalizeURI(forUri));
			}

			// comment (allow deletion of an obsolete acl)

			/*
			 * try { resourceNode =
			 * NodeUtil.getLocalHome().getByPath(normalizeAbsoluteURIPath(resource)); }
			 * catch (ObjectNotFoundException e) { throw new
			 * URINotExistException(normalizeURI(resource)); }
			 */
			HashMap acls = forUriNode.getAcls();
			URI del = null;

			for (Iterator i = acls.keySet().iterator(); i.hasNext();) {
				URI key = (URI) i.next();

				if (key.getPath().equals(normalizeURI(resource).getPath())) {
					del = key;
				}
			}

			acls.remove(del);
			forUriNode.setAcls(acls);
			forUriNode.setUpdateDate(new Date());

			// recursive
			if (recursive) {
				Collection children = forUriNode.getChildren2();

				for (Iterator i = children.iterator(); i.hasNext();) {
					NodeLocal child = (NodeLocal) i.next();
					deleteAcl(child.getURI(), resource, true);
				}
			}

			// event
			Libresource.throwEvent(new LibresourceEvent(forUri, KernelConstants.SERVICE, "Node", getConnectedResource(), KernelConstants.EVENT_DELETE_ACLS));
		} catch (LibresourceSecurityException se) {
			ctx.setRollbackOnly();
			throw se;
		} catch (URINotExistException se) {
			ctx.setRollbackOnly();
			throw se;
		} catch (Exception e) {
			ctx.setRollbackOnly();
			throw new KernelServiceException("Error in deleteAcl : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public AclValue[] getAcls(URI uri) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			NodeLocal uriNode;

			try {
				uriNode = NodeUtil.getLocalHome().getByPath(normalizeAbsoluteURIPath(uri));
			} catch (ObjectNotFoundException e) {
				throw new URINotExistException(normalizeURI(uri));
			}

			HashMap acls = uriNode.getAcls();
			AclValue[] result = new AclValue[acls.size()];
			int k = 0;

			for (Iterator i = acls.keySet().iterator(); i.hasNext();) {
				URI key = (URI) i.next();
				result[k] = new AclValue(key, (String[]) acls.get(key));
				result[k].setOwnerUri(normalizeURI(result[k].getOwnerUri()));
				k++;
			}

			return result;
		} catch (LibresourceSecurityException se) {
			throw se;
		} catch (URINotExistException se) {
			throw se;
		} catch (Exception e) {
			throw new KernelServiceException("Error in getAcls : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public String[] listKernelPermissions() {
		return new String[] { KernelConstants.SECURITY_READ, KernelConstants.SECURITY_CREATE, KernelConstants.SECURITY_UPDATE, KernelConstants.SECURITY_DELETE };
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public String[] listKernelEvents() {
		return new String[] { KernelConstants.EVENT_CREATE, KernelConstants.EVENT_BIND, KernelConstants.EVENT_UNBIND, KernelConstants.EVENT_MOVE_TO,
				KernelConstants.EVENT_MOVE_FROM, KernelConstants.EVENT_DELETE, KernelConstants.EVENT_CHOWN, KernelConstants.EVENT_CREATE_ACLS,
				KernelConstants.EVENT_RESET_ACLS, KernelConstants.EVENT_DELETE_ACLS, KernelConstants.EVENT_SET_PROPERTY };
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public boolean checkIfOwner(URI user, URI uri) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			String connectedUser = normalizeAbsoluteURIPath(user);

			try {
				NodeLocal uriNode = NodeUtil.getLocalHome().getByPath(normalizeAbsoluteURIPath(uri));
				LibresourceResourceIdentifier ownerResourceIdentifier;

				if (uriNode.getOwner().trim().length() == 0) {
					return false;

					// ownerResourceIdentifier = null;
				} else {
					URI ownerUri = NodeUtil.getLocalHome().findByPrimaryKey(uriNode.getOwner()).getURI();
					ownerResourceIdentifier = systemLookup(normalizeAbsoluteURIPath(ownerUri));
				}

				LibresourceResourceIdentifier connectedResourceIdentifier = systemLookup(connectedUser);

				if (connectedResourceIdentifier == null) {
					return false;
				}

				if (ownerResourceIdentifier != null) {
					LibresourceService libresourceService = Libresource.getService(connectedResourceIdentifier.getService());
					LibresourceService libresourceService2 = Libresource.getService(ownerResourceIdentifier.getService());

					if (libresourceService.canMapResourceForSecurity(new URI(connectedUser), NodeUtil.getLocalHome().findByPrimaryKey(uriNode.getOwner())
							.getURI())
							|| libresourceService2.canMapResourceForSecurity(new URI(connectedUser), NodeUtil.getLocalHome().findByPrimaryKey(
									uriNode.getOwner()).getURI())) {
						return true;
					}
				} else {
					LibresourceService libresourceService = Libresource.getService(connectedResourceIdentifier.getService());

					if (libresourceService.canMapResourceForSecurity(new URI(connectedUser), NodeUtil.getLocalHome().findByPrimaryKey(uriNode.getOwner())
							.getURI())) {
						return true;
					}
				}

				return false;
			} catch (ObjectNotFoundException e) {
				return false;

				// throw new URINotExistException(normalizeURI(uri));
			}
		} catch (LibresourceSecurityException se) {
			throw se;
		} catch (URINotExistException se) {
			throw se;
		} catch (Exception e) {
			throw new KernelServiceException("Error in checkIfOwner : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public boolean checkIfOwner(URI uri) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			if (!exist(uri)) {
				throw new URINotExistException(normalizeURI(uri));
			}

			if (isSuperUser()) {
				return true;
			}

			return checkIfOwner(getConnectedResource(), uri);
		} catch (LibresourceSecurityException se) {
			throw se;
		} catch (URINotExistException se) {
			throw se;
		} catch (Exception e) {
			throw new KernelServiceException("Error in checkIfOwner : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public URI getConnectedResource() throws KernelServiceException, LibresourceSecurityException {
		try {
			if (principals.get((isAuthentified()) ? ctx.getCallerPrincipal().getName() : ANONYMOUS) != null) {
				URI principal = normalizeURI((URI) (principals.get((isAuthentified()) ? ctx.getCallerPrincipal().getName() : ANONYMOUS)));

				return principal;
			}

			if (principalMapper == null) {
				StringTokenizer tokenizer = new StringTokenizer((String) new InitialContext().lookup("java:comp/env/principal-mapper-options"), ";");
				Hashtable mapperOptions = new Hashtable();

				while (tokenizer.hasMoreTokens()) {
					String option = tokenizer.nextToken();
					mapperOptions.put(option.substring(0, option.indexOf("=")).trim(), option.substring(option.indexOf("=") + 1).trim());
				}

				principalMapper = (PrincipalMapper) Class.forName((String) new InitialContext().lookup("java:comp/env/principal-mapper")).getConstructor(
						new Class[] { Map.class }).newInstance(new Object[] { mapperOptions });
			}

			URI resource = principalMapper.map((isAuthentified()) ? ctx.getCallerPrincipal() : null);
			principals.put((isAuthentified()) ? ctx.getCallerPrincipal().getName() : ANONYMOUS, resource);

			return normalizeURI(resource);
		} catch (Exception e) {
			throw new KernelServiceException("Error in getConnectedResource : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public boolean isAuthentified() throws KernelServiceException, LibresourceSecurityException {
		try {
			return !(ctx.getCallerPrincipal().getName().equals("client-login") || ctx.getCallerPrincipal().getName().equals("JOnAS_client")
					|| ctx.getCallerPrincipal().getName().equals("jonas") || ctx.getCallerPrincipal().getName().equals("ANONYMOUS"));
		} catch (IllegalStateException e) {
			// no securityContext
			return false;
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public boolean isSuperUser() throws KernelServiceException {
		try {
			return ctx.getCallerPrincipal().getName().equals("Administrateur");
		} catch (IllegalStateException e) {
			// no securityContext
			return false;
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public int countResources(String service, String resourceType) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			return NodeUtil.getLocalHome().countResource(service, resourceType);
		} catch (Exception e) {
			throw new KernelServiceException("Error in countResource : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public LibresourceResourceValue[] listResourcesAt(URI atUri, String service, String resourceType) throws KernelServiceException,
			LibresourceSecurityException, URINotExistException {
		try {
			NodeLocal pathNodeLocal;

			try {
				pathNodeLocal = NodeUtil.getLocalHome().getByPath(normalizeAbsoluteURIPath(atUri));
			} catch (ObjectNotFoundException e) {
				throw new URINotExistException(normalizeURI(atUri));
			}

			if (!checkSecurity(atUri, KernelConstants.SECURITY_READ)) {
				return new LibresourceResourceValue[0];
			}

			Collection collection = NodeUtil.getLocalHome().findResources(pathNodeLocal.getId(), service + "/" + resourceType + "/%");
			Vector filteredFor_R_RightUris = new Vector();
			// Optimization
			HashMap acls = null;
			Boolean cacheResult = null;
			boolean okToKeep = false;
			HashMap canMapLocalCache = new HashMap();
			URI connectedUserURI = getConnectedResource();
			canMapLocalCache.put(connectedUserURI.getPath(), Boolean.TRUE);
			MembershipService membershipService = (MembershipService) Libresource.getService(MembershipConstants.SERVICE);
			boolean rootUser = isSuperUser();

			for (Iterator i = collection.iterator(); i.hasNext();) {
				NodeLocal uriNode = (NodeLocal) i.next();
				acls = uriNode.getAcls();
				okToKeep = false;

				if (!rootUser)
					for (Iterator aclKey = acls.keySet().iterator(); aclKey.hasNext() && !okToKeep;) {
						URI key = (URI) aclKey.next();
						if (contains((String[]) acls.get(key), KernelConstants.SECURITY_READ)) {
							cacheResult = (Boolean) canMapLocalCache.get(key.getPath());
							if (cacheResult == null) {
								try {
									if (!key.getPath().startsWith("/users/") && membershipService.canMapResourceForSecurity(connectedUserURI, key)) {
										canMapLocalCache.put(key.getPath(), Boolean.TRUE);
										okToKeep = true;
									} else {
										canMapLocalCache.put(key.getPath(), Boolean.FALSE);
									}
								} catch (Exception e) {
									canMapLocalCache.put(key.getPath(), Boolean.FALSE);
								}
							} else {
								okToKeep = cacheResult.booleanValue();
							}
						}
					}

				if (okToKeep || rootUser) {
					LibresourceResourceValue libresourceResourceValue = new LibresourceResourceValue();
					libresourceResourceValue.setUri(normalizeURI(uriNode.getURI()));
					LibresourceResourceIdentifier resourceIdentifier = LibresourceResourceIdentifier.deserialize(uriNode.getBindedResourceIdentifier());
					libresourceResourceValue.setShortResourceName(uriNode.getShortName());
					libresourceResourceValue.setCreationDate(uriNode.getCreationDate());
					libresourceResourceValue.setUpdateDate(uriNode.getUpdateDate());
					libresourceResourceValue.setOwner(new URI(uriNode.getOwner()));
					libresourceResourceValue.setLibresourceResourceIdentifier(resourceIdentifier);
					filteredFor_R_RightUris.add(libresourceResourceValue);
				}
			}

			// sort the list by ShortResourceName
			filteredFor_R_RightUris = sortResourcesList(filteredFor_R_RightUris);

			LibresourceResourceValue[] resourceValues = new LibresourceResourceValue[filteredFor_R_RightUris.size()];

			for (int i = 0; i < resourceValues.length; i++) {
				resourceValues[i] = (LibresourceResourceValue) filteredFor_R_RightUris.get(i);
			}

			return resourceValues;
		} catch (URINotExistException se) {
			throw se;
		} catch (Exception e) {
			throw new KernelServiceException("Error in listResourcesAt : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public LibresourceResourceValue[] listResourcesAt(URI atUri) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		long time = System.currentTimeMillis();
		try {
			NodeLocal pathNodeLocal;

			try {
				pathNodeLocal = NodeUtil.getLocalHome().getByPath(normalizeAbsoluteURIPath(atUri));
			} catch (ObjectNotFoundException e) {
				throw new URINotExistException(normalizeURI(atUri));
			}

			if (!checkSecurity(atUri, KernelConstants.SECURITY_READ)) {
				return new LibresourceResourceValue[0];
			}

			Collection collection = pathNodeLocal.getChildren2();
			// Collection collection =
			// NodeUtil.getLocalHome().findChildren(pathNodeLocal.getId());

			Vector filteredFor_R_RightUris = new Vector();
			Vector resources = new Vector();
			Vector empty = new Vector();
			HashMap acls = null;
			Boolean cacheResult = null;
			boolean okToKeep = false;
			HashMap canMapLocalCache = new HashMap();
			URI connectedUserURI = getConnectedResource();
			canMapLocalCache.put(connectedUserURI.getPath(), Boolean.TRUE);
			MembershipService membershipService = (MembershipService) Libresource.getService(MembershipConstants.SERVICE);
			boolean rootUser = isSuperUser();

			for (Iterator i = collection.iterator(); i.hasNext();) {
				NodeLocal uriNode = (NodeLocal) i.next();
				acls = uriNode.getAcls();
				okToKeep = false;

				if (!rootUser)
					for (Iterator aclKey = acls.keySet().iterator(); aclKey.hasNext() && !okToKeep;) {
						URI key = (URI) aclKey.next();
						if (contains((String[]) acls.get(key), KernelConstants.SECURITY_READ)) {
							cacheResult = (Boolean) canMapLocalCache.get(key.getPath());
							if (cacheResult == null) {
								try {
									if (!key.getPath().startsWith("/users/") && membershipService.canMapResourceForSecurity(connectedUserURI, key)) {
										canMapLocalCache.put(key.getPath(), Boolean.TRUE);
										okToKeep = true;
									} else {
										canMapLocalCache.put(key.getPath(), Boolean.FALSE);
									}
								} catch (Exception e) {
									canMapLocalCache.put(key.getPath(), Boolean.FALSE);
								}
							} else {
								okToKeep = cacheResult.booleanValue();
							}
						}
					}

				if (okToKeep || rootUser) {
					LibresourceResourceValue libresourceResourceValue = new LibresourceResourceValue();
					libresourceResourceValue.setUri(normalizeURI(uriNode.getURI()));

					if (uriNode.getBindedResourceIdentifier() != null) {
						LibresourceResourceIdentifier resourceIdentifier = LibresourceResourceIdentifier.deserialize(uriNode.getBindedResourceIdentifier());
						libresourceResourceValue.setShortResourceName(uriNode.getShortName());
						libresourceResourceValue.setLibresourceResourceIdentifier(resourceIdentifier);
						libresourceResourceValue.setCreationDate(uriNode.getCreationDate());
						libresourceResourceValue.setUpdateDate(uriNode.getUpdateDate());
						libresourceResourceValue.setOwner(new URI(uriNode.getOwner()));
						resources.add(libresourceResourceValue);
					} else {
						libresourceResourceValue.setIsEmpty(true);
						libresourceResourceValue.setShortResourceName("/" + uriNode.getPathPart() + " (Empty Node)");
						libresourceResourceValue.setCreationDate(uriNode.getCreationDate());
						libresourceResourceValue.setUpdateDate(uriNode.getUpdateDate());
						libresourceResourceValue.setOwner(new URI(uriNode.getOwner()));
						empty.add(libresourceResourceValue);
					}
				}
			}
			// sort the list by ShortResourceName
			filteredFor_R_RightUris = sortResourcesList(resources);
			filteredFor_R_RightUris.addAll(empty);

			LibresourceResourceValue[] resourceValues = new LibresourceResourceValue[filteredFor_R_RightUris.size()];

			for (int i = 0; i < resourceValues.length; i++) {
				resourceValues[i] = (LibresourceResourceValue) filteredFor_R_RightUris.get(i);
			}

			return resourceValues;
		} catch (URINotExistException se) {
			throw se;
		} catch (Exception e) {
			e.printStackTrace();
			throw new KernelServiceException("Error in listResourcesAt : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public LibresourceResourceValue getResource(URI atUri) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			LibresourceResourceValue libresourceResourceValue = new LibresourceResourceValue();
			libresourceResourceValue.setUri(normalizeURI(atUri));

			LibresourceResourceIdentifier resourceIdentifier = systemLookup(atUri);

			if (resourceIdentifier != null) {
				libresourceResourceValue.setShortResourceName(getShortName(normalizeURI(atUri))); // resourceIdentifier));
				libresourceResourceValue.setLibresourceResourceIdentifier(resourceIdentifier);
			} else {
				libresourceResourceValue.setIsEmpty(true);
				libresourceResourceValue.setShortResourceName(libresourceResourceValue.getUri().toString().substring(
						libresourceResourceValue.getUri().toString().lastIndexOf("/")));
			}

			NodeLocal node = NodeUtil.getLocalHome().getByPath(normalizeAbsoluteURIPath(atUri));
			libresourceResourceValue.setCreationDate(node.getCreationDate());
			libresourceResourceValue.setUpdateDate(node.getUpdateDate());
			libresourceResourceValue.setOwner(getOwner(node.getURI()));

			return libresourceResourceValue;
		} catch (URINotExistException se) {
			throw se;
		} catch (LibresourceSecurityException se) {
			throw se;
		} catch (Exception e) {
			throw new KernelServiceException("Error in getResource : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public Vector getNavigationBar(URI atUri) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			NodeLocal pathNodeLocal;

			try {
				pathNodeLocal = NodeUtil.getLocalHome().getByPath(normalizeAbsoluteURIPath(atUri));
			} catch (ObjectNotFoundException e) {
				throw new URINotExistException(normalizeURI(atUri));
			}

			Vector bar = new Vector();

			while (pathNodeLocal != null) {
				String[] node = new String[3];
				node[0] = normalizeURI(pathNodeLocal.getURI()).getPath();
				node[1] = pathNodeLocal.getPathPart();

				if (pathNodeLocal.getBindedResourceIdentifier() != null) {
					LibresourceResourceIdentifier resourceIdentifier = LibresourceResourceIdentifier.deserialize(pathNodeLocal.getBindedResourceIdentifier());
					node[2] = Libresource.getShortName(normalizeURI(pathNodeLocal.getURI())); // resourceIdentifier);
				} else {
					node[2] = pathNodeLocal.getPathPart();
				}

				bar.add(node);

				if (pathNodeLocal.getParent().trim().length() == 0) {
					pathNodeLocal = null;
				} else {
					pathNodeLocal = NodeUtil.getLocalHome().findByPrimaryKey(pathNodeLocal.getParent());
				}
			}

			Collections.reverse(bar);

			return bar;
		} catch (URINotExistException se) {
			throw se;
		} catch (Exception e) {
			throw new KernelServiceException("Error in getNavigationBar : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public URI[] listChildren(URI atUri) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			NodeLocal pathNodeLocal;

			try {
				pathNodeLocal = NodeUtil.getLocalHome().getByPath(normalizeAbsoluteURIPath(atUri));
			} catch (ObjectNotFoundException e) {
				throw new URINotExistException(normalizeURI(atUri));
			}

			if (!checkSecurity(atUri, KernelConstants.SECURITY_READ)) {
				return new URI[0];
			}

			Collection collection = pathNodeLocal.getChildren2();
			Vector filteredFor_R_RightUris = new Vector();

			for (Iterator i = collection.iterator(); i.hasNext();) {
				URI uri = ((NodeLocal) i.next()).getURI();

				// if (checkSecurity(uri, "Kernel:READ")) {
				filteredFor_R_RightUris.add(uri);

				// }
			}

			// sort the uris list
			filteredFor_R_RightUris = sortUrisList(filteredFor_R_RightUris);

			URI[] uris = new URI[filteredFor_R_RightUris.size()];

			for (int i = 0; i < uris.length; i++) {
				uris[i] = (URI) filteredFor_R_RightUris.get(i);
			}

			return uris;
		} catch (URINotExistException se) {
			throw se;
		} catch (Exception e) {
			throw new KernelServiceException("Error in listChildren : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public boolean hasChildren(URI atUri) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			NodeLocal pathNodeLocal;

			try {
				pathNodeLocal = NodeUtil.getLocalHome().getByPath(normalizeAbsoluteURIPath(atUri));
			} catch (ObjectNotFoundException e) {
				throw new URINotExistException(normalizeURI(atUri));
			}

			return pathNodeLocal.getChildren().isEmpty();
		} catch (URINotExistException se) {
			throw se;
		} catch (Exception e) {
			throw new KernelServiceException("Error in hasChildren : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public boolean checkSecurity(URI user, URI uri, String permission) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			String connectedResource = normalizeAbsoluteURIPath(user);

			// owner has all permissions
			if (checkIfOwner(user, uri)) {
				return true;
			}

			/*
			 * if (permission.equals(KernelConstants.SECURITY_KERNEL_READ)) { if
			 * (connectedResource.equals(normalizeAbsoluteURIPath(uri))) {
			 * return true; } }
			 */

			// Check before the Kernel:Update right if its a specific right
			if (!permission.startsWith(KernelConstants.SECURITY) && checkSecurity(user, uri, KernelConstants.SECURITY_UPDATE)) {
				return true;
			}

			try {
				NodeLocal uriNode = NodeUtil.getLocalHome().getByPath(normalizeAbsoluteURIPath(uri));
				HashMap acls = uriNode.getAcls();

				for (Iterator i = acls.keySet().iterator(); i.hasNext();) {
					URI key = (URI) i.next();

					// TODO remove obsolete acls...
					if (exist(key)) {
						LibresourceResourceIdentifier ownerResourceIdentifier = systemLookup(normalizeAbsoluteURIPath(key));
						LibresourceResourceIdentifier connectedResourceIdentifier = systemLookup(connectedResource);

						if (connectedResourceIdentifier == null) {
							return false;
						}

						LibresourceService libresourceService = Libresource.getService(connectedResourceIdentifier.getService());
						LibresourceService libresourceService2 = Libresource.getService(ownerResourceIdentifier.getService());

						if (libresourceService.canMapResourceForSecurity(new URI(connectedResource), new URI(normalizeAbsoluteURIPath(key)))
								|| libresourceService2.canMapResourceForSecurity(new URI(connectedResource), new URI(normalizeAbsoluteURIPath(key)))) {
							if (contains((String[]) acls.get(key), permission)) {
								return true;
							}
						}
					}
				}

				return false;
			} catch (ObjectNotFoundException e) {
				return false;

				// throw new URINotExistException(normalizeURI(uri));
			}
		} catch (URINotExistException se) {
			throw se;
		} catch (LibresourceSecurityException se) {
			throw se;
		} catch (Exception e) {
			throw new KernelServiceException("Error in checkSecurity : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public boolean checkSecurity(URI uri, String permission) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			if (!exist(uri)) {
				throw new URINotExistException(normalizeURI(uri));
			}

			if (isSuperUser()) {
				return true;
			}

			return checkSecurity(getConnectedResource(), uri, permission);
		} catch (URINotExistException se) {
			return false;
		} catch (LibresourceSecurityException se) {
			throw se;
		} catch (Exception e) {
			throw new KernelServiceException("Error in checkSecurity : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public boolean zubZum(URI connectedUri, URI toUri) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			if (isSuperUser()) {
				return true;
			}

			LibresourceResourceIdentifier connectedResource = systemLookup(normalizeAbsoluteURIPath(connectedUri));
			LibresourceResourceIdentifier toResource = null;

			try {
				toResource = systemLookup(normalizeAbsoluteURIPath(toUri));
			} catch (Exception e) {
				return false;
			}

			LibresourceService connectedResourceService = Libresource.getService(connectedResource.getService());
			LibresourceService toResourceService = Libresource.getService(toResource.getService());

			return connectedResourceService
					.canMapResourceForSecurity(new URI(normalizeAbsoluteURIPath(connectedUri)), new URI(normalizeAbsoluteURIPath(toUri)))
					|| toResourceService.canMapResourceForSecurity(new URI(normalizeAbsoluteURIPath(connectedUri)), new URI(normalizeAbsoluteURIPath(toUri)));
		} catch (URINotExistException se) {
			throw se;
		} catch (LibresourceSecurityException se) {
			throw se;
		} catch (Exception e) {
			throw new KernelServiceException("Error in zubZum : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public boolean zubZum(URI toUri) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		return zubZum(getConnectedResource(), toUri);
	}

	/**
	 * Seb coding to prevent stack overflow
	 * 
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public void indexAll() throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			Collection nodes = NodeUtil.getLocalHome().findAll();
			NodeLocal currentNode = null;

			for (Iterator i = nodes.iterator(); i.hasNext();) {
				currentNode = (NodeLocal) i.next();

				if (currentNode.hasResourceBinded()) {
					System.out.println("Indexing: " + currentNode.getURI());

					try {
						Libresource.index(currentNode.getURI());
						System.out.println(" + OK");
					} catch (RuntimeException e) {
						System.out.println(" - KO");
					}
				} else {
					System.out.println("NOT Indexing: " + currentNode.getURI());
				}
			}
		} catch (Exception e) {
			throw new KernelServiceException("Error in indexAll : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public void flushIndexBase() throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			LibresourceSearch.getInstance().flush();
		} catch (Exception e) {
			throw new KernelServiceException("Error in flushIndexBase : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public LibresourceSearchResult[] search(String query) throws KernelServiceException, LibresourceSecurityException {
		return search(null, null, null, query);
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public String highlight(String text, String quey) throws KernelServiceException, LibresourceSecurityException {
		try {
			return LibresourceSearch.getInstance().highlight(text, quey);
		} catch (LibresourceException e) {
			throw new KernelServiceException(e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public LibresourceSearchResult[] search(String uriPattern, String service, String resource, String query) throws KernelServiceException,
			LibresourceSecurityException {
		try {
			LibresourceSearchResult[] results = LibresourceSearch.getInstance().search(uriPattern, service, resource, query);
			Vector filteredResults = new Vector();

			for (int i = 0; i < results.length; i++) {
				try {
					if (checkSecurity(results[i].getUri(), KernelConstants.SECURITY_READ)) {
						filteredResults.add(results[i]);
					}
				} catch (Exception e) {
					//
				}
			}

			LibresourceSearchResult[] finalResults = new LibresourceSearchResult[filteredResults.size()];

			for (int i = 0; i < filteredResults.size(); i++) {
				finalResults[i] = (LibresourceSearchResult) filteredResults.get(i);
				finalResults[i].setResourceIdentifier(systemLookup(normalizeAbsoluteURIPath(finalResults[i].getUri())));
			}

			return finalResults;
		} catch (Exception e) {
			throw new KernelServiceException("Error in search(" + query + ") : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public Object getPropertyObject(URI uri, String key) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		return getPropertyObject(uri, key, false);
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public String getProperty(URI uri, String key) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		return getProperty(uri, key, false);
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public URI getNodeDefiningProperty(URI uri, String key) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			NodeLocal node = null;

			try {
				node = NodeUtil.getLocalHome().getByPath(normalizeAbsoluteURIPath(uri));
			} catch (ObjectNotFoundException e) {
				throw new URINotExistException(normalizeURI(uri));
			}

			while (node != null) {
				Object result = getPropertyObject(node, key);

				if (result != null) {
					return normalizeURI(node.getURI());
				}

				if ((node.getParent() == null) || (node.getParent().trim().length() == 0)) {
					node = null;
				} else {
					node = NodeUtil.getLocalHome().findByPrimaryKey(node.getParent());
				}
			}

			return null;
		} catch (URINotExistException se) {
			throw se;
		} catch (LibresourceSecurityException se) {
			throw se;
		} catch (Exception e) {
			throw new KernelServiceException("Error in getNodeDefiningProperty : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public String getProperty(URI uri, String key, boolean inherit) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		return (String) getPropertyObject(uri, key, inherit);
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public Object getPropertyObject(URI uri, String key, boolean inherit) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			if (!exist(uri)) {
				return null;
			}

			if (!checkSecurity(uri, KernelConstants.SECURITY_READ)) {
				return null;
			}

			try {
				NodeLocal node = NodeUtil.getLocalHome().getByPath(normalizeAbsoluteURIPath(uri));

				if (!inherit) {
					return getPropertyObject(node, key);
				} else {
					URI parent = getNodeDefiningProperty(uri, key);

					if (parent != null) {
						return getPropertyObject(parent, key, false);
					}
				}

				return null;
			} catch (ObjectNotFoundException e) {
				throw new URINotExistException(normalizeURI(uri));
			}
		} catch (URINotExistException se) {
			throw se;
		} catch (LibresourceSecurityException se) {
			throw se;
		} catch (Exception e) {
			throw new KernelServiceException("Error in getProperty : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public HashMap listPropertiesAtNode(URI uri) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			if (!checkSecurity(uri, KernelConstants.SECURITY_READ)) {
				throw new LibresourceSecurityException(uri, KernelConstants.SECURITY_READ);
			}

			try {
				NodeLocal node = NodeUtil.getLocalHome().getByPath(normalizeAbsoluteURIPath(uri));

				HashMap map = node.getProperties();

				if (map == null) {
					return new HashMap();
				}

				return map;
			} catch (ObjectNotFoundException e) {
				throw new URINotExistException(normalizeURI(uri));
			}
		} catch (URINotExistException se) {
			throw se;
		} catch (LibresourceSecurityException se) {
			throw se;
		} catch (Exception e) {
			throw new KernelServiceException("Error in listPropertiesAtNode : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Required"
	 */
	public void setProperty(URI uri, String key, Object value) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			if (!checkSecurity(uri, KernelConstants.SECURITY_UPDATE)) {
				throw new LibresourceSecurityException(uri, KernelConstants.SECURITY_UPDATE);
			}

			systemSetProperty(uri, key, value);
		} catch (URINotExistException se) {
			ctx.setRollbackOnly();
			throw se;
		} catch (LibresourceSecurityException se) {
			ctx.setRollbackOnly();
			throw se;
		} catch (Exception e) {
			ctx.setRollbackOnly();
			throw new KernelServiceException("Error in setProperty : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Required"
	 */
	public void systemSetProperty(URI uri, String key, Object value) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			try {
				NodeLocal node = NodeUtil.getLocalHome().getByPath(normalizeAbsoluteURIPath(uri));
				HashMap map = node.getProperties();

				if (map == null) {
					map = new HashMap();
				}

				map.put(key, value);
				node.setProperties(map);
				node.setUpdateDate(new Date());

				// event
				Libresource.throwEvent(new LibresourceEvent(uri, KernelConstants.SERVICE, "Node", getConnectedResource(), KernelConstants.EVENT_SET_PROPERTY,
						key + " = " + value));
			} catch (ObjectNotFoundException e) {
				throw new URINotExistException(normalizeURI(uri));
			}
		} catch (URINotExistException se) {
			ctx.setRollbackOnly();
			throw se;
		} catch (LibresourceSecurityException se) {
			ctx.setRollbackOnly();
			throw se;
		} catch (Exception e) {
			ctx.setRollbackOnly();
			throw new KernelServiceException("Error in systemSetProperty : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Required"
	 */
	public void setCreationDate(URI uri, Date date) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			if (!checkSecurity(uri, KernelConstants.SECURITY_UPDATE)) {
				throw new LibresourceSecurityException(uri, KernelConstants.SECURITY_UPDATE);
			}

			NodeLocal node = NodeUtil.getLocalHome().getByPath(normalizeAbsoluteURIPath(uri));
			node.setCreationDate(date);
		} catch (URINotExistException se) {
			ctx.setRollbackOnly();
			throw se;
		} catch (LibresourceSecurityException se) {
			ctx.setRollbackOnly();
			throw se;
		} catch (Exception e) {
			ctx.setRollbackOnly();
			throw new KernelServiceException("Error in setCreationDate : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public Date getCreationDate(URI uri) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			if (!checkSecurity(uri, KernelConstants.SECURITY_READ)) {
				throw new LibresourceSecurityException(uri, KernelConstants.SECURITY_READ);
			}

			NodeLocal node = NodeUtil.getLocalHome().getByPath(normalizeAbsoluteURIPath(uri));

			return node.getCreationDate();
		} catch (URINotExistException se) {
			ctx.setRollbackOnly();
			throw se;
		} catch (LibresourceSecurityException se) {
			ctx.setRollbackOnly();
			throw se;
		} catch (Exception e) {
			ctx.setRollbackOnly();
			throw new KernelServiceException("Error in setCreationDate : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Required"
	 */
	public void setUpdateDate(URI uri, Date date) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			// Shouldn't be needed

			/*
			 * if (!checkSecurity(uri, KernelConstants.SECURITY_KERNEL_READ)) {
			 * throw new LibresourceSecurityException(uri,
			 * KernelConstants.SECURITY_KERNEL_READ); }
			 */
			NodeLocal node = NodeUtil.getLocalHome().getByPath(normalizeAbsoluteURIPath(uri));
			node.setUpdateDate(date);
		} catch (URINotExistException se) {
			ctx.setRollbackOnly();
			throw se;
		} catch (LibresourceSecurityException se) {
			ctx.setRollbackOnly();
			throw se;
		} catch (Exception e) {
			ctx.setRollbackOnly();
			throw new KernelServiceException("Error in setUpdateDate : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public Date getUpdateDate(URI uri) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			if (!checkSecurity(uri, KernelConstants.SECURITY_UPDATE)) {
				throw new LibresourceSecurityException(uri, KernelConstants.SECURITY_UPDATE);
			}

			NodeLocal node = NodeUtil.getLocalHome().getByPath(normalizeAbsoluteURIPath(uri));

			return node.getUpdateDate();
		} catch (URINotExistException se) {
			ctx.setRollbackOnly();
			throw se;
		} catch (LibresourceSecurityException se) {
			ctx.setRollbackOnly();
			throw se;
		} catch (Exception e) {
			ctx.setRollbackOnly();
			throw new KernelServiceException("Error in getUpdateDate : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Required"
	 */
	public void resetAcls(URI forUri) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			if (!checkIfCanChangeRights(forUri)) {
				throw new LibresourceSecurityException(forUri, "OWNER or CRUD");
			}

			systemResetAcls(forUri);

			// event
			Libresource.throwEvent(new LibresourceEvent(forUri, KernelConstants.SERVICE, "Node", getConnectedResource(), KernelConstants.EVENT_RESET_ACLS));
		} catch (LibresourceSecurityException se) {
			ctx.setRollbackOnly();
			throw se;
		} catch (URINotExistException se) {
			ctx.setRollbackOnly();
			throw se;
		} catch (Exception e) {
			ctx.setRollbackOnly();
			throw new KernelServiceException("Error in systemResetAcls : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public String normalizeAbsoluteURIPath(URI uri) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			// String thisHost = ((uri.getHost() == null) ? "" : uri.getHost())
			// + ((uri.getPort() == -1) ? "" : (":" + uri.getPort()));
			//
			// if (((uri.getScheme() != null) &&
			// !uri.getScheme().equals(getUriScheme())) || (!thisHost.equals("")
			// && !thisHost.equals(getUriHost()))) {
			// throw new URINotExistException("Invalid URI for this site : " +
			// uri + " [this site scheme is " + getUriScheme() + "; this site
			// host is " +
			// getUriHost() + "]");
			// }
			String path = uri.normalize().getPath();

			if (!path.startsWith("/")) {
				path = "/" + path;
			}

			if (path.endsWith("/")) {
				path = path.substring(0, path.length() - 1);
			}

			if (path.equals("")) {
				return "/";
			}

			if (path.startsWith("/..")) {
				throw new URINotExistException("Invalid URI");
			}

			return path;
		} catch (URINotExistException se) {
			throw se;
		} catch (Exception e) {
			throw new KernelServiceException("Error in normalizeAbsoluteURIPath : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Required"
	 */
	public void createSymbolicLink(URI uri, URI toUri) throws KernelServiceException, LibresourceSecurityException {
		try {
			SymbolicLinkResourceLocal linkResourceLocal = SymbolicLinkResourceUtil.getLocalHome().create(normalizeAbsoluteURIPath(toUri));
			bind(linkResourceLocal.getLibresourceResourceIdentifier(), uri, getShortName(toUri));

			// event
			LibresourceEvent event = new LibresourceEvent(uri, linkResourceLocal.getLibresourceResourceIdentifier(), getConnectedResource(),
					"libresourceCore.symbolicLink.create");
			Libresource.throwEvent(event);
		} catch (LibresourceSecurityException se) {
			ctx.setRollbackOnly();
			throw se;
		} catch (Exception e) {
			ctx.setRollbackOnly();
			throw new KernelServiceException("Error in createSymbolicLink : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Required"
	 */
	public void editSymbolicLink(URI uri, URI toUri) throws KernelServiceException, LibresourceSecurityException {
		try {
			LibresourceResourceIdentifier id = lookup(uri);
			SymbolicLinkResourceLocal linkResourceLocal = (SymbolicLinkResourceLocal) Libresource.findResource(id, SymbolicLinkResourceLocal.class);
			linkResourceLocal.setLinkURI(normalizeAbsoluteURIPath(toUri));

			try {
				setShortName(uri, Libresource.getShortName(toUri));
			} catch (Exception e) {
				setShortName(uri, "Invalid Symbolic Link to " + toUri);
			}

			setUpdateDate(uri, new Date());

			// event
			LibresourceEvent event = new LibresourceEvent(uri, linkResourceLocal.getLibresourceResourceIdentifier(), getConnectedResource(),
					"libresourceCore.symbolicLink.edit");
			Libresource.throwEvent(event);
		} catch (LibresourceSecurityException se) {
			ctx.setRollbackOnly();
			throw se;
		} catch (Exception e) {
			ctx.setRollbackOnly();
			throw new KernelServiceException("Error in editSymbolicLink : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public SymbolicLinkResourceValue getSymbolicLink(URI uri) throws KernelServiceException, LibresourceSecurityException {
		try {
			LibresourceResourceIdentifier id = lookup(uri);
			SymbolicLinkResourceLocal linkResourceLocal = (SymbolicLinkResourceLocal) Libresource.findResource(id, SymbolicLinkResourceLocal.class);
			SymbolicLinkResourceValue linkResourceValue = linkResourceLocal.getSymbolicLinkResourceValue();
			linkResourceValue.setUri(normalizeURI(uri));

			return linkResourceValue;
		} catch (LibresourceSecurityException se) {
			throw se;
		} catch (Exception e) {
			throw new KernelServiceException("Error in getSymbolicLink : " + e.getMessage(), e);
		}
	}

	// system

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public LibresourceResourceIdentifier systemLookup(String path) throws Exception {
		try {
			NodeLocal pathNodeLocal = NodeUtil.getLocalHome().getByPath(path);

			return LibresourceResourceIdentifier.deserialize(pathNodeLocal.getBindedResourceIdentifier());
		} catch (ObjectNotFoundException e) {
			throw new URINotExistException(new URI(path));
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Supports"
	 */
	public LibresourceResourceIdentifier systemLookup(URI uri) throws Exception {
		try {
			NodeLocal pathNodeLocal = NodeUtil.getLocalHome().getByPath(normalizeAbsoluteURIPath(uri));

			return LibresourceResourceIdentifier.deserialize(pathNodeLocal.getBindedResourceIdentifier());
		} catch (ObjectNotFoundException e) {
			throw new URINotExistException(normalizeURI(uri));
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Required"
	 */
	public void systemResetAcls(URI forUri) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			NodeLocal forUriNode;

			try {
				forUriNode = NodeUtil.getLocalHome().getByPath(normalizeAbsoluteURIPath(forUri));
			} catch (ObjectNotFoundException e) {
				throw new URINotExistException(normalizeURI(forUri));
			}

			forUriNode.setAcls(new HashMap());
			forUriNode.setUpdateDate(new Date());

			// event
			Libresource.throwEvent(new LibresourceEvent(forUri, KernelConstants.SERVICE, "Node", getConnectedResource(), KernelConstants.EVENT_RESET_ACLS));
		} catch (LibresourceSecurityException se) {
			ctx.setRollbackOnly();
			throw se;
		} catch (URINotExistException se) {
			ctx.setRollbackOnly();
			throw se;
		} catch (Exception e) {
			ctx.setRollbackOnly();
			throw new KernelServiceException("Error in systemResetAcls : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Required"
	 */
	public void systemChown(URI uri, URI owner, boolean recursive) throws KernelServiceException, LibresourceSecurityException, URINotExistException {
		try {
			NodeLocal uriNode;
			NodeLocal ownerNode;

			try {
				uriNode = NodeUtil.getLocalHome().getByPath(normalizeAbsoluteURIPath(uri));
			} catch (ObjectNotFoundException e) {
				throw new URINotExistException(normalizeURI(uri));
			}

			try {
				ownerNode = NodeUtil.getLocalHome().getByPath(normalizeAbsoluteURIPath(owner));
			} catch (ObjectNotFoundException e) {
				throw new URINotExistException(normalizeURI(owner));
			}

			uriNode.setOwner(ownerNode.getId());

			if (recursive) {
				Collection children = uriNode.getChildren2();

				for (Iterator i = children.iterator(); i.hasNext();) {
					NodeLocal child = (NodeLocal) i.next();
					chown(child.getURI(), owner, true);
				}
			}

			setUpdateDate(uri, new Date());

			// event
			Libresource.throwEvent(new LibresourceEvent(uri, KernelConstants.SERVICE, "Node", getConnectedResource(), KernelConstants.EVENT_CHOWN));
		} catch (LibresourceSecurityException se) {
			ctx.setRollbackOnly();
			throw se;
		} catch (URINotExistException se) {
			ctx.setRollbackOnly();
			throw se;
		} catch (Exception e) {
			ctx.setRollbackOnly();
			throw new KernelServiceException("Error in systemChown : " + e.getMessage(), e);
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Required"
	 */
	public void systemCreateURI(URI uri) throws Exception {
		try {
			NodeLocalHome pathNodeLocalHome = NodeUtil.getLocalHome();
			StringTokenizer tokenizer = new StringTokenizer(normalizeAbsoluteURIPath(uri), "/");
			ArrayList pathParts = new ArrayList();

			while (tokenizer.hasMoreTokens()) {
				pathParts.add(tokenizer.nextToken());
			}

			NodeLocal tempParent = null;

			try {
				tempParent = pathNodeLocalHome.findByPrimaryKey("666");
			} catch (ObjectNotFoundException e) {
				tempParent = pathNodeLocalHome.create(null, "");
			}

			for (Iterator i = pathParts.iterator(); i.hasNext();) {
				String pathPart = (String) i.next();

				// tempPath.append(pathPart);
				String childid = (String) tempParent.getChildren().get(pathPart);

				if (childid == null) {
					NodeLocal newNode = pathNodeLocalHome.create(tempParent, pathPart);
					copyAcls(tempParent, newNode);
					tempParent.addChild(newNode);
					tempParent = newNode;
				} else {
					tempParent = pathNodeLocalHome.findByPrimaryKey(childid);
				}
			}

			// event
			// Libresource.throwEvent(new LibresourceEvent(uri,
			// KernelConstants.SERVICE,
			// "Node", getConnectedResource(), "kernel.create"));
		} catch (Exception e) {
			ctx.setRollbackOnly();
			throw e;
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Required"
	 */
	public void systemCopyAcl(URI from, URI to) throws Exception {
		try {
			NodeLocalHome pathNodeLocalHome = NodeUtil.getLocalHome();
			copyAcls(pathNodeLocalHome.getByPath(normalizeAbsoluteURIPath(from)), pathNodeLocalHome.getByPath(normalizeAbsoluteURIPath(to)));
		} catch (Exception e) {
			ctx.setRollbackOnly();
			throw e;
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Required"
	 */
	public void systemDeleteChildren(URI uri) throws Exception {
		try {
			NodeLocalHome pathNodeLocalHome = NodeUtil.getLocalHome();
			NodeLocal nodeLocal = pathNodeLocalHome.getByPath(normalizeAbsoluteURIPath(uri));

			for (Iterator i = nodeLocal.getChildren().values().iterator(); i.hasNext();) {
				NodeUtil.getLocalHome().findByPrimaryKey((String) i.next()).remove();
			}
		} catch (Exception e) {
			ctx.setRollbackOnly();
			throw e;
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Required"
	 */
	public void systemCreateAcl(URI resource, URI owner, String[] permissions) throws Exception {
		try {
			NodeLocal node = NodeUtil.getLocalHome().getByPath(normalizeAbsoluteURIPath(resource));
			HashMap acls = node.getAcls();

			if (acls == null) {
				acls = new HashMap();
			}

			// Keep only owner that have at least on right
			String[] resultingPermission = filterPermissions(LibresourceResourceIdentifier.deserialize(node.getBindedResourceIdentifier()), permissions);

			if (resultingPermission.length > 0) {
				acls.put(owner, resultingPermission);
			}

			node.setAcls(acls);
			node.setUpdateDate(new Date());

			// event
			Libresource.throwEvent(new LibresourceEvent(resource, KernelConstants.SERVICE, "Node", getConnectedResource(), KernelConstants.EVENT_CREATE_ACLS));
		} catch (Exception e) {
			ctx.setRollbackOnly();
			throw e;
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Required"
	 */
	public void systemBind(LibresourceResourceIdentifier resourceIdentifier, URI uri, String resourceName) throws Exception {
		try {
			systemCreateURI(uri);

			NodeLocalHome pathNodeLocalHome = NodeUtil.getLocalHome();
			NodeLocal node = pathNodeLocalHome.getByPath(normalizeAbsoluteURIPath(uri));
			node.setBindedResourceIdentifier(resourceIdentifier.serialize());
			node.setShortName(resourceName);
			node.setUpdateDate(new Date());

			// event
			Libresource.throwEvent(new LibresourceEvent(uri, resourceIdentifier, getConnectedResource(), KernelConstants.EVENT_BIND));
		} catch (Exception e) {
			ctx.setRollbackOnly();
			throw e;
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="RequiresNew"
	 */
	public void systemBootstrap() throws Exception {
		try {
			NodeLocalHome nodeLocalHome = NodeUtil.getLocalHome();

			// create the root node
			NodeLocal root = nodeLocalHome.create(null, "");

			// create the guest user
			MembershipService membershipService = (MembershipService) Libresource.getService(MembershipConstants.SERVICE);
			ProfileResourceLocal profileResourceLocal = ProfileResourceUtil.getLocalHome().create(membershipService.getUnauthentifiedUserId(), "Guest", "", "",
					new HashMap());
			URI profileUri = new URI(membershipService.getUsersRootURI().getPath() + "/" + membershipService.getUnauthentifiedUserId());
			systemCreateURI(profileUri);
			systemBind(profileResourceLocal.getLibresourceResourceIdentifier(), profileUri, "Guest");
			systemCreateAcl(profileUri, profileUri, new String[] { KernelConstants.SECURITY_READ });
		} catch (Exception e) {
			ctx.setRollbackOnly();
			throw e;
		}
	}

	/**
	 * @ejb.interface-method
	 * @ejb.transaction type="Required"
	 */
	public void systemInitConfig() throws Exception {
		URI configUri = getServerConfigURI();
		systemCreateURI(configUri);
		systemResetAcls(configUri);

		MembershipService membershipService = (MembershipService) Libresource.getService(MembershipConstants.SERVICE);
		systemCreateAcl(configUri, new URI(membershipService.getUsersRootURI().getPath() + "/" + membershipService.getUnauthentifiedUserId()),
				new String[] { KernelConstants.SECURITY_READ });
		systemCreateAcl(configUri, new URI(membershipService.getUsersRootURI().getPath() + "/root"), new String[] { KernelConstants.SECURITY_READ,
				KernelConstants.SECURITY_UPDATE });
		systemSetProperty(configUri, "register-users", membershipService.getUsersRootURI().getPath() + "/" + membershipService.getUnauthentifiedUserId());
	}

	// helpers
	private Vector sortResourcesList(Vector resourcesList) throws Exception {
		Collections.sort(resourcesList, new Comparator() {
			public int compare(Object o1, Object o2) {
				if (!(o1 instanceof LibresourceResourceValue)) {
					return 0;
				}

				if ((((LibresourceResourceValue) o1).getShortResourceName() != null) && (((LibresourceResourceValue) o2).getShortResourceName() != null)) {
					// System.out.println(((LibresourceResourceValue)
					// o1).getShortResourceName());
					// System.out.println(((LibresourceResourceValue)
					// o2).getShortResourceName());
					// System.out.println(((LibresourceResourceValue)
					// o1).getShortResourceName()).compareToIgnoreCase(((LibresourceResourceValue)
					// o2).getShortResourceName()
					return (((LibresourceResourceValue) o1).getShortResourceName()).compareToIgnoreCase(((LibresourceResourceValue) o2).getShortResourceName());
				}

				return 0;
			}
		});

		return resourcesList;
	}

	private Vector sortUrisList(Vector urisList) throws Exception {
		Collections.sort(urisList, new Comparator() {
			public int compare(Object o1, Object o2) {
				if (!(o1 instanceof URI)) {
					return 0;
				}

				return (((URI) o1).compareTo((URI) o2));
			}
		});

		return urisList;
	}

	private void copyAcls(NodeLocal fromUri, NodeLocal toUri) throws Exception {
		if (fromUri == null) {
			return;
		}

		LibresourceResourceIdentifier resourceIdentifier = LibresourceResourceIdentifier.deserialize(toUri.getBindedResourceIdentifier());

		HashMap acls = new HashMap();
		HashMap parentAcls = fromUri.getAcls();

		if (parentAcls != null) {
			for (Iterator i = parentAcls.keySet().iterator(); i.hasNext();) {
				URI key = (URI) i.next();
				String[] permissions = (String[]) parentAcls.get(key);
				acls.put(key, filterPermissions(resourceIdentifier, permissions));
			}
		}

		toUri.setAcls(acls);
		toUri.setUpdateDate(new Date());

		// event
		Libresource
				.throwEvent(new LibresourceEvent(toUri.getURI(), KernelConstants.SERVICE, "Node", getConnectedResource(), KernelConstants.EVENT_CREATE_ACLS));
	}

	private String[] filterPermissions(LibresourceResourceIdentifier resourceIdentifier, String[] permissions) throws Exception {
		String[] availablesPermissions = Libresource.listAvailablesPermissions(resourceIdentifier);
		Vector filtered = new Vector();

		for (int i = 0; i < permissions.length; i++) {
			if (permissions[i] == null) {
				continue;
			}

			String permission = permissions[i].trim();

			if (permission.startsWith(KernelConstants.SERVICE) || contains(availablesPermissions, permission)) {
				filtered.add(permission);
			}
		}

		String[] result = new String[filtered.size()];

		for (int i = 0; i < result.length; i++) {
			result[i] = (String) filtered.get(i);
		}

		return result;
	}

	private boolean contains(String[] list, String element) {
		if ((list == null) || (element == null)) {
			return false;
		}

		for (int i = 0; i < list.length; i++) {
			if (list[i] == null) {
				continue;
			}

			if (list[i].equals(element)) {
				return true;
			}
		}

		return false;
	}

	public Object getPropertyObject(NodeLocal node, String key) throws Exception {
		if (node == null) {
			return null;
		}

		HashMap properties = node.getProperties();

		if (properties == null) {
			return null;
		}

		return properties.get(key);
	}

	public String getProperty(NodeLocal node, String key) throws Exception {
		return (String) getPropertyObject(node, key);
	}
}
