/*
 * Copyright (c) 2005, Adam Pordzik. See LICENSE file for details
 */

package org.libresource.membership;


import java.io.IOException;
import javax.naming.ldap.ExtendedRequest;
import javax.naming.ldap.ExtendedResponse;
import javax.naming.NamingException;
import com.sun.jndi.ldap.Ber;
import com.sun.jndi.ldap.BerEncoder;

/**
 * This class implements an LDAPv3 Extended Request for password
 * changing as defined in <a href="http://www.ietf.org/rfc/rfc3062.txt">RFC
 * 3062 - LDAP Password Modify Extended Operation</a>.
 * <p>
 * The object identifier for this operation is "1.3.6.1.4.1.4203.1.11.1". If
 * any of userIdentity, oldPasswd or newPasswd value is defined, a extended
 * request sequence is assembled. <p>
 * <tt>PasswordChangeRequest</tt>/<tt>PasswordChangeResponse</tt> are used to
 * change passwords of directory entries which can hold userPassword, as defined
 * e.g. for person, inetOrgPerson, simpleSecurityObject and so on. Whichever
 * password encryption method is used to store a password (hash) value in this
 * attribute is determined by server itself. (It's not even guaranteed that
 * the attribute containing a user's password is named userPassword)
 * <p>
 * PasswordChangeRequest takes an user identity, usually a DN, the old password
 * and the new password as parameters. Every of these values can be left out by
 * setting it to null. Depending on directory server implementation, it is
 * commonly sufficient only to provide the new password, since user's identity
 * and his authorization for changing is already granted by bind operation thru
 * its credentials. However a directory adminstrator, such as "cn=Manager,dc=..."
 * is able to change anothers password by simply specifying its DN.
 * <p>
 * Specifing old password normally is never required, but it may be used if one
 * want to change foreign passwords without having administrative rights (Never
 * tried this, but I think such would be only depends on server's ACL settings.)
 * <p>
 * Omitting new password mentioning will ask the server to generate a ramdom
 * password which will be replied to PasswordChangeResponse.
 * <p>
 * <b>Example:</b>
 * <pre>
 * // Open an LDAP association
 * Hashtable env = new Hashtable();
 * env.put(Context.SECURITY_PRINCIPAL, "cn=manager,dc=example,dc=org");
 * env.put(Context.SECURITY_CREDENTIALS, "secret");
 * LdapContext ctx = new InitialLdapContext();
 *
 * String userDn = "cn=Some User,dc=example,dc=org";
 * String oldPwd = null;
 * String userDn = "newsecret"; // null for auto generated password
 *
 * // Perform a password change extended operation
 * PasswordChangeResponse pcr = (PasswordChangeResponse) ctx.extendedOperation (
 *              new PasswordChangeRequest (userDn, newPwd, oldPwn));
 *
 * //  pcr.getGenPassword() holds generated password
 * //  pcr.getNewPassword() holds in either case the new password
 *
 * // Close the LDAP association
 * ctx.close;
 * </pre></blockquote>
 * @author Adam Pordzik
 * @see ChangePasswordResponse
 *
 */

public class PasswordChangeRequest implements ExtendedRequest {

        static final String OID = "1.3.6.1.4.1.4203.1.11.1";

        String userDn = null;
        String oldPwd = null;
        String newPwd = null;

    /**
     * Constructs an empty password changing extended request for currently
     * binded user identity which generates a new, arbitrary password.
     */

        public PasswordChangeRequest () {

        }

    /**
     * Constructs a password changing extended request with optional parameters.
     *
     * @param userDn    Omitting userDn (userDn=null) relates operation to
     *                                  currently binded identity. directory administrators with
     *                                  write permission to users userPassword attribute can
     *                                  provide other's identities
     * @param oldPwd    Normally there's no need providing old password, since
     *                                  password change extended request should not operate on
     *                                  unauthenticated binds.
     * @param newPwd    The new password for userDn. Omitting newPwd (newPwd=null)
     *                                  calls server on generating an arbitrarily password, which
     *                                  is returned by PasswordChangeResponse
     */

        public PasswordChangeRequest (String userDn, String oldPwd, String newPwd) {
                this.userDn = userDn;
                this.oldPwd = oldPwd;
                this.newPwd = newPwd;
        }

        /**
         * Creates the password change extended response object. Passes also newPwd
         * for PasswordChangeResponse always having a valid value for getNewPassword()
         * <p>
         * Creates an extended response object that corresponds to the
     * LDAP change password extended request.
     * <p>
     * The result must be a concrete subclass of PasswordChangeResponse
     *
     * @param oid      The object identifier of the extended response.
     *                  Password response does not return an OID
     * @param berValue The possibly null ASN.1 BER encoded value of the
     *                  extended response. This is the raw BER bytes
     *                  including the tag (an initiatory sequence) and length
     *                  of the response value. (the response does not include
     *                  an OID.) Its value may contain an generated password,
     *                  but response is not expected to contain any value if
     *                  newPwd is specified.
     * @param offset   The starting position in berValue of the bytes to use.
     *                  Usually zero.
     * @param length   The number of bytes in berValue to use.
     * @return                  The password changing extended response object.
     * @exception               NamingException If a naming exception was encountered.
         * @see PasswordChangeResponse
     */

        public ExtendedResponse createExtendedResponse (String oid, byte[] berValue, int offset, int length)
        throws NamingException {
                // ggf. altes Password mitgeben.
                return new PasswordChangeResponse (oid, berValue, offset, length, newPwd);
        }

        /**
         * Retrieves the password change request's object identifier string.
         *
         * @return The object identifier string, "1.3.6.1.4.1.4203.1.11.1".
         */

        public String getID() {
                return OID;
        }

        /**
     * Retrieves the password change request's ASN.1 BER encoded value.
     * <p>
     * If any of userDn, oldPwd or newPwd is not null, getEncodedValue()
     * assebles an ASN.1 sequence containing all non-null values as optional
     * fiels in an ASN.1 context.
     * @return The ASN.1 BER encoded request result, if no one parameter
     * is spcified null is returned.
     * @throws IllegalStateException Throws exception if values cannot get
     * BER encoded.
     * <p>
     */

        public byte[] getEncodedValue() {

                // No requestValues present, requesting without PasswdModifyRequestValue-sequence
                if (userDn == null && newPwd == null && oldPwd == null)
                        return null;

                BerEncoder berEncoder = new BerEncoder();
                try {
                        berEncoder.beginSeq (Ber.ASN_SEQUENCE | Ber.ASN_CONSTRUCTOR); // = 0x30;
                        // Tagged, optional context fields
                        if (userDn != null)
                        berEncoder.encodeString (userDn, Ber.ASN_CONTEXT | 0x00, false); // User DN
                        if (oldPwd != null)
                        berEncoder.encodeString (oldPwd, Ber.ASN_CONTEXT | 0x01, true); // old PW
                        if (newPwd != null)
                        berEncoder.encodeString (newPwd, Ber.ASN_CONTEXT | 0x02, true); // new PW
                        berEncoder.endSeq ();

                } catch (IOException encodingException) {
                        throw new IllegalStateException ("BER encoding error: " + encodingException.getMessage());

                }
//              BerEncoder.dumpBER(System.out, "Request:\n", berEncoder.getBuf(), 0, berEncoder.getDataLen());
                return berEncoder.getTrimmedBuf();
        }

        static final long serialVersionUID = 4126595370571669865L;

} // PasswordChangeRequest

