/*
 * Copyright (c) 2005, Adam Pordzik. See LICENSE file for details
 */

package org.libresource.membership;

import java.io.IOException;
import com.sun.jndi.ldap.Ber;
import com.sun.jndi.ldap.BerDecoder;
import javax.naming.ldap.ExtendedResponse;
import javax.naming.NamingException;

/**
 * Corresponding response class for PasswordChangeRequest. It mainly implements
 * public methods getGenPassword() and getNewPassword() for accessing these
 * values. getGenPassword may return null if no password was generated, whereas
 * getNewPassword always contains user's new password, either generated or the
 * one formerly provided by request operation.
 * <p>
 * ChangePasswordResponse has no sensible public contructor, since
 * PasswordChangePesponse Objects are about only to be created as a result
 * of a preceding PasswordChangeRequest.
 *
 * @see ChangePasswordRequest
 * @author Adam Pordzik
 *
 */
public class PasswordChangeResponse implements ExtendedResponse {

        static final String requestOID = "1.3.6.1.4.1.4203.1.11.1";

        private String responseOID = null;
        private byte[] responseBer = null;
        private BerDecoder berDecoder;
        private String newPwd = null;
        private String genPwd = null;

        /**
         * PasswordChangeResponse constructor is invoked by
         * PasswordChangeRequest#createExtendedResponse()with common BER response
         * values and newPwd, which might be null. If so, PasswordChangeResponse
         * assumes that a new password is generated by the directory server.
         *
         * @param oid
         * @param berValue
         * @param offset
         * @param length
         * @param newPwd
         * @see PasswordChangeRequest#createExtendedResponse(String, byte[], int, int);
         * @throws NamingException
         */

        // Soll das public sein, oder reicht es aus nur im package-scope sichtbar zu sein?

        PasswordChangeResponse (String oid, byte berValue[], int offset, int length, String newPwd)
        throws NamingException {
                this.responseOID = oid; // PwdExop never returns an OID
                this.responseBer = berValue;
                this.berDecoder = new BerDecoder (berValue, offset, length);
                this.newPwd = newPwd;
        }

        /**
         *
         * @return Returns either the generated password or the supplied password
         * @throws NamingException Throws Exception if error occurs during BER decoding
         * of generated password.
         */

        public String getNewPassword ()
        throws NamingException {
                String retPwd = getGenPassword();
                if (retPwd == null) {
                        return newPwd;
                }
                return retPwd;
        }

        /**
         *
         * @return new password or null if no password was generated.
         * @throws NamingException Throws Exception if Error occurs during BER decoding.
         */

        public String getGenPassword ()
        throws NamingException {
                // Keine Antwort ist auch eine Antwort
                if (responseBer == null) return null;

                BerDecoder berDecoder = new BerDecoder (responseBer, 0, responseBer.length);
//              BerDecoder.dumpBER(System.out, "Response:\n", responseBer, 0, responseBer.length);
                try {
                        /* Kept this unused stuff as a hint for next time, since com.sun.jndi.ldap.Ber*
                         * are poorly documented. Next time? Well,
                         * http://www.ietf.org/internet-drafts/draft-behera-ldap-password-policy-09.txt
                         * has to be implemeted when it's finished... ;-)
                        int[] sizes = new int[2];
                        sizes[1] = 7;
                        System.out.println(berDecoder.peekByte());
                        int tag = berDecoder.parseSeq (sizes); // Why does this return an empty array?
                        System.out.println("Sequence has " +sizes.length + " Elemnets. First one is of "+ sizes[1] + " Bytes");
                        genPwd = berDecoder.parseStringWithTag (Ber.ASN_CONTEXT | 0x00, true, null);
                        */
                        berDecoder.parseSeq (null);
                        genPwd = berDecoder.parseStringWithTag (Ber.ASN_CONTEXT | 0x00, true, null);

                } catch (IOException e) {
                        throw new NamingException ("BER decoding error: " + e.getMessage());

                }
                return genPwd;
        }

        /**
         * @return The password change request operation OID. Resonse does not contain
         * an OID, so it shoul be always null.
         */

        public String getID() {
                return this.responseOID;
        }

        /**
         * @return Raw DER encoded response, may be null
         */

        public byte[] getEncodedValue() {
                return this.responseBer;
        }

        static final long serialVersionUID = 2214037158278153551L;
}
