/**
 * LibreSource
 * Copyright (C) 2004-2008 Artenum SARL / INRIA
 * http://www.libresource.org - contact@artenum.com
 *
 * This file is part of the LibreSource software, 
 * which can be used and distributed under license conditions.
 * The license conditions are provided in the LICENSE.TXT file 
 * at the root path of the packaging that enclose this file. 
 * More information can be found at 
 * - http://dev.libresource.org/home/license
 *
 * Initial authors :
 *
 * Guillaume Bort / INRIA
 * Francois Charoy / Universite Nancy 2
 * Julien Forest / Artenum
 * Claude Godart / Universite Henry Poincare
 * Florent Jouille / INRIA
 * Sebastien Jourdain / INRIA / Artenum
 * Yves Lerumeur / Artenum
 * Pascal Molli / Universite Henry Poincare
 * Gerald Oster / INRIA
 * Mariarosa Penzi / Artenum
 * Gerard Sookahet / Artenum
 * Raphael Tani / INRIA
 *
 * Contributors :
 *
 * Stephane Bagnier / Artenum
 * Amadou Dia / Artenum-IUP Blois
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 */
package org.libresource.membership.ejb;

import org.libresource.Libresource;
import org.libresource.LibresourceEvent;
import org.libresource.LibresourceException;
import org.libresource.LibresourceResourceIdentifier;
import org.libresource.LibresourceResourceValue;
import org.libresource.LibresourceServiceBase;

import org.libresource.kernel.KernelConstants;
import org.libresource.kernel.LibresourceSecurityException;
import org.libresource.kernel.URINotExistException;
import org.libresource.kernel.UriManager;
import org.libresource.kernel.interfaces.KernelService;

import org.libresource.membership.PasswordChangeRequest;
import org.libresource.membership.PasswordChangeResponse;
import org.libresource.membership.GroupExportHandler;
import org.libresource.membership.GroupImportHandler;
import org.libresource.membership.MembershipConstants;
import org.libresource.membership.MembershipServiceException;
import org.libresource.membership.ProfileExportHandler;
import org.libresource.membership.ProfileImportHandler;
import org.libresource.membership.ejb.model.GroupResourceValue;
import org.libresource.membership.ejb.model.LDAPGroupResourceValue;
import org.libresource.membership.ejb.model.ProfileResourceValue;
import org.libresource.membership.interfaces.GroupResourceLocal;
import org.libresource.membership.interfaces.LDAPGroupResourceLocal;
import org.libresource.membership.interfaces.ProfileResourceLocal;
import org.libresource.membership.util.GroupResourceUtil;
import org.libresource.membership.util.LDAPGroupResourceUtil;
import org.libresource.membership.util.ProfileResourceUtil;

import org.libresource.search.LibresourceIndexableContent;

import org.libresource.xml.LibresourceExportHandler;
import org.libresource.xml.LibresourceImportHandler;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.Writer;

import java.net.URI;

import java.security.MessageDigest;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.Statement;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;

import javax.naming.AuthenticationException;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;
import javax.naming.directory.Attribute;
import javax.naming.directory.Attributes;
import javax.naming.directory.BasicAttributes;
import javax.naming.directory.DirContext;
import javax.naming.directory.InitialDirContext;
import javax.naming.directory.SearchControls;
import javax.naming.directory.SearchResult;
import javax.naming.ldap.LdapContext;
import javax.naming.ldap.InitialLdapContext;


/**
 * LibreSource
 *
 * @author <a href="mailto:bort@loria.fr">Guillaume Bort</a> - <a
 *         href="http://www.inria.fr">INRIA Lorraine</a>
 *
 * The LibreSource Default Membership Service
 *
 * @libresource.service name="Membership" depends="Kernel"
 *
 * @ejb.env-entry name="users-provider-1"
 *                value="org.libresource.membership.providers.InternalBaseUsersProvider"
 * @ejb.env-entry name="options-users-provider-1" value=""
 * @ejb.env-entry name="users-provider-2"
 *                value="org.libresource.membership.providers.GuestUserProvider"
 * @ejb.env-entry name="options-users-provider-2" value=""
 * @ejb.env-entry name="users-root-uri" value="/users"
 * @ejb.env-entry name="unauthentified-userId" value="guest"
 * @ejb.env-entry name="allAuthentifiedUsers-group-uri" value="/groups/all"
 *
 */
public abstract class MembershipServiceBean extends LibresourceServiceBase {
    private static transient URI usersRoot;
    private static transient String unauthentifieduserid;
    private static transient URI allAuthentifiedUsersGroupURI;

    /**
     * @ejb.interface-method
     * @ejb.transaction type="Supports"
     */
    public URI getUsersRootURI() throws MembershipServiceException, LibresourceSecurityException {
        try {
            KernelService kernelService = (KernelService) Libresource.getService(KernelConstants.SERVICE);

            if (usersRoot == null) {
                usersRoot = kernelService.normalizeURI(new URI((String) new InitialContext().lookup("java:comp/env/users-root-uri")));
            }

            return usersRoot;
        } catch (Exception e) {
            throw new MembershipServiceException(e);
        }
    }

    /**
     * @ejb.interface-method
     * @ejb.transaction type="Supports"
     */
    public String getUnauthentifiedUserId() throws MembershipServiceException, LibresourceSecurityException {
        try {
            if (unauthentifieduserid == null) {
                unauthentifieduserid = (String) new InitialContext().lookup("java:comp/env/unauthentified-userId");
            }

            return unauthentifieduserid;
        } catch (Exception e) {
            throw new MembershipServiceException(e);
        }
    }

    /**
     * @ejb.interface-method
     * @ejb.transaction type="Supports"
     */
    public URI getAllAuthentifiedUserGroupURI() throws MembershipServiceException, LibresourceSecurityException {
        try {
            if (allAuthentifiedUsersGroupURI == null) {
                allAuthentifiedUsersGroupURI = new URI((String) new InitialContext().lookup("java:comp/env/allAuthentifiedUsers-group-uri"));
            }

            return allAuthentifiedUsersGroupURI;
        } catch (Exception e) {
            throw new MembershipServiceException(e);
        }
    }

    /**
     * @ejb.interface-method
     * @ejb.transaction type="Required"
     */
    public void createGroup(URI uri, String name, String description)
        throws MembershipServiceException, LibresourceSecurityException {
        try {
            GroupResourceLocal groupResourceLocal = GroupResourceUtil.getLocalHome().create(name, description);
            KernelService kernelService = (KernelService) Libresource.getService(KernelConstants.SERVICE);
            kernelService.bind(groupResourceLocal.getLibresourceResourceIdentifier(), uri, name);

            // event
            LibresourceEvent event = new LibresourceEvent(uri, groupResourceLocal.getLibresourceResourceIdentifier(), kernelService.getConnectedResource(),
                    MembershipConstants.EVENT_GROUP_CREATE);
            Libresource.throwEvent(event);
        } catch (LibresourceSecurityException se) {
            ctx.setRollbackOnly();
            throw se;
        } catch (Exception e) {
            ctx.setRollbackOnly();
            throw new MembershipServiceException("Error in createGroup : " + e.getMessage(), e);
        }
    }

    /**
     * @ejb.interface-method
     * @ejb.transaction type="Required"
     */
    public void createLDAPGroup(URI uri, String name, String description)
        throws MembershipServiceException, LibresourceSecurityException {
        try {
            LDAPGroupResourceLocal ldapGroupResourceLocal = LDAPGroupResourceUtil.getLocalHome().create(name, description);
            KernelService kernelService = (KernelService) Libresource.getService(KernelConstants.SERVICE);
            kernelService.bind(ldapGroupResourceLocal.getLibresourceResourceIdentifier(), uri, name);

            // event
            LibresourceEvent event = new LibresourceEvent(uri, ldapGroupResourceLocal.getLibresourceResourceIdentifier(), kernelService.getConnectedResource(),
                    MembershipConstants.EVENT_GROUP_CREATE);
            Libresource.throwEvent(event);
        } catch (LibresourceSecurityException se) {
            ctx.setRollbackOnly();
            throw se;
        } catch (Exception e) {
            ctx.setRollbackOnly();
            throw new MembershipServiceException("Error in createGroup : " + e.getMessage(), e);
        }
    }

    /**
     * @ejb.interface-method
     * @ejb.transaction type="Required"
     */
    public void createProfile(String userId, String userName, String userEmail, String userJabberId, HashMap userInfos)
        throws MembershipServiceException, LibresourceSecurityException {
        try {
            if (!userId.matches("[A-Za-z0-9]+")) {
                throw new Exception("invalid user id (userId should match [A-Za-z0-9])");
            }

            ProfileResourceLocal profileResourceLocal = ProfileResourceUtil.getLocalHome().create(userId, userName, userEmail, userJabberId, userInfos);
            KernelService kernelService = (KernelService) Libresource.getService(KernelConstants.SERVICE);

            URI profileUri = new URI(getUsersRootURI().getPath() + "/" + userId);

            // TODO system ???
            kernelService.bind(profileResourceLocal.getLibresourceResourceIdentifier(), profileUri, userName);
            kernelService.systemCopyAcl(getUsersRootURI(), profileUri);
            kernelService.systemCreateAcl(profileUri, profileUri, new String[] { KernelConstants.SECURITY_READ, KernelConstants.SECURITY_UPDATE });

            // event
            LibresourceEvent event = new LibresourceEvent(profileUri, profileResourceLocal.getLibresourceResourceIdentifier(),
                    kernelService.getConnectedResource(), MembershipConstants.EVENT_PROFILE_CREATE);
            Libresource.throwEvent(event);
        } catch (LibresourceSecurityException se) {
            ctx.setRollbackOnly();
            throw se;
        } catch (Exception e) {
            e.printStackTrace();
            ctx.setRollbackOnly();
            throw new MembershipServiceException("Error in createProfile : " + e.getMessage(), e);
        }
    }

    /**
     * @ejb.interface-method
     * @ejb.transaction type="Required"
     */
    public void editProfile(URI uri, String userName, String userEmail, String userJabberId)
        throws MembershipServiceException, LibresourceSecurityException {
        try {
            KernelService kernelService = (KernelService) Libresource.getService(KernelConstants.SERVICE);
            LibresourceResourceIdentifier resourceIdentifier = kernelService.lookup(uri);

            if (!kernelService.checkSecurity(uri, KernelConstants.SECURITY_UPDATE)) {
                throw new LibresourceSecurityException(uri, KernelConstants.SECURITY_UPDATE);
            }

            ProfileResourceLocal profile = (ProfileResourceLocal) Libresource.findResource(resourceIdentifier, ProfileResourceLocal.class);
            profile.setFullName(userName);
            profile.setEmail(userEmail);
            profile.setJabberId(userJabberId);
            profile.setAccountStatus(profile.getAccountStatus());
            profile.setOnlineStatus(profile.getOnlineStatus());

            // update node
            kernelService.setShortName(uri, userName);
            kernelService.setUpdateDate(uri, new Date());

            // event
            LibresourceEvent event = new LibresourceEvent(uri, resourceIdentifier, kernelService.getConnectedResource(), MembershipConstants.EVENT_PROFILE_EDIT);
            Libresource.throwEvent(event);

            // indexation
            Libresource.index(uri);
        } catch (LibresourceSecurityException se) {
            ctx.setRollbackOnly();
            throw se;
        } catch (Exception e) {
            ctx.setRollbackOnly();
            throw new MembershipServiceException("Error in editProfile : " + e.getMessage(), e);
        }
    }

    /**
     * @ejb.interface-method
     * @ejb.transaction type="Required"
     */
    public void changeProfileStatus(URI uri, int newStatus)
        throws MembershipServiceException, LibresourceSecurityException {
        try {
            KernelService kernelService = (KernelService) Libresource.getService(KernelConstants.SERVICE);
            LibresourceResourceIdentifier resourceIdentifier = kernelService.lookup(uri);

            if (!kernelService.isSuperUser()) {
                throw new LibresourceSecurityException(uri, KernelConstants.SECURITY_UPDATE);
            }

            ProfileResourceLocal profile = (ProfileResourceLocal) Libresource.findResource(resourceIdentifier, ProfileResourceLocal.class);
            profile.setAccountStatus(newStatus);

            // event
            LibresourceEvent event = new LibresourceEvent(uri, resourceIdentifier, kernelService.getConnectedResource(), MembershipConstants.EVENT_PROFILE_EDIT);
            Libresource.throwEvent(event);

            // indexation
            Libresource.index(uri);
        } catch (LibresourceSecurityException se) {
            ctx.setRollbackOnly();
            throw se;
        } catch (Exception e) {
            ctx.setRollbackOnly();
            throw new MembershipServiceException("Error in changeProfileStatus : " + e.getMessage(), e);
        }
    }

    /**
     * @ejb.interface-method
     * @ejb.transaction type="Required"
     */
    public void setUserInfo(URI uri, String key, Object value)
        throws MembershipServiceException, LibresourceSecurityException {
        try {
            KernelService kernelService = (KernelService) Libresource.getService(KernelConstants.SERVICE);
            LibresourceResourceIdentifier resourceIdentifier = kernelService.lookup(uri);

            if (!kernelService.checkSecurity(uri, KernelConstants.SECURITY_UPDATE)) {
                throw new LibresourceSecurityException(uri, KernelConstants.SECURITY_UPDATE);
            }

            ProfileResourceLocal profile = (ProfileResourceLocal) Libresource.findResource(resourceIdentifier, ProfileResourceLocal.class);
            profile.setInfo(key, value);

            // update node
            kernelService.setUpdateDate(uri, new Date());

            // event
            LibresourceEvent event = new LibresourceEvent(uri, resourceIdentifier, kernelService.getConnectedResource(), MembershipConstants.EVENT_PROFILE_EDIT);
            Libresource.throwEvent(event);
        } catch (LibresourceSecurityException se) {
            ctx.setRollbackOnly();
            throw se;
        } catch (Exception e) {
            ctx.setRollbackOnly();
            throw new MembershipServiceException("Error in setUserInfo : " + e.getMessage(), e);
        }
    }

    /**
     * @ejb.interface-method
     * @ejb.transaction type="Required"
     */
    public void removeUserInfo(URI uri, String key) throws MembershipServiceException, LibresourceSecurityException {
        try {
            KernelService kernelService = (KernelService) Libresource.getService(KernelConstants.SERVICE);
            LibresourceResourceIdentifier resourceIdentifier = kernelService.lookup(uri);

            if (!kernelService.checkSecurity(uri, KernelConstants.SECURITY_UPDATE)) {
                throw new LibresourceSecurityException(uri, KernelConstants.SECURITY_UPDATE);
            }

            ProfileResourceLocal profile = (ProfileResourceLocal) Libresource.findResource(resourceIdentifier, ProfileResourceLocal.class);
            profile.removeInfo(key);

            // update node
            kernelService.setUpdateDate(uri, new Date());

            // event
            LibresourceEvent event = new LibresourceEvent(uri, resourceIdentifier, kernelService.getConnectedResource(), MembershipConstants.EVENT_PROFILE_EDIT);
            Libresource.throwEvent(event);
        } catch (LibresourceSecurityException se) {
            ctx.setRollbackOnly();
            throw se;
        } catch (Exception e) {
            ctx.setRollbackOnly();
            throw new MembershipServiceException("Error in removeUserInfo : " + e.getMessage(), e);
        }
    }

    /**
     * @ejb.interface-method
     * @ejb.transaction type="Required"
     */
    public void editGroup(URI uri, String name, String description)
        throws MembershipServiceException, LibresourceSecurityException, URINotExistException {
        try {
            KernelService kernelService = (KernelService) Libresource.getService(KernelConstants.SERVICE);
            LibresourceResourceIdentifier resourceIdentifier = kernelService.lookup(uri);

            if (!kernelService.checkSecurity(uri, KernelConstants.SECURITY_UPDATE)) {
                throw new LibresourceSecurityException(uri, KernelConstants.SECURITY_UPDATE);
            }

            GroupResourceLocal group = (GroupResourceLocal) Libresource.findResource(resourceIdentifier, GroupResourceLocal.class);
            group.setName(name);
            group.setDescription(description);

            // update node
            kernelService.setShortName(uri, name);
            kernelService.setUpdateDate(uri, new Date());

            // event
            LibresourceEvent event = new LibresourceEvent(uri, resourceIdentifier, kernelService.getConnectedResource(), MembershipConstants.EVENT_GROUP_EDIT);
            Libresource.throwEvent(event);

            // indexation
            Libresource.index(uri);
        } catch (LibresourceSecurityException se) {
            ctx.setRollbackOnly();
            throw se;
        } catch (URINotExistException se) {
            ctx.setRollbackOnly();
            throw se;
        } catch (Exception e) {
            ctx.setRollbackOnly();
            throw new MembershipServiceException("Error in editGroup : " + e.getMessage(), e);
        }
    }

    /**
     * @ejb.interface-method
     * @ejb.transaction type="Required"
     */
    public void editLDAPGroup(URI uri, String name, String description, String serverURL, String ldapGroupDefinition, String field, String userID)
        throws MembershipServiceException, LibresourceSecurityException, URINotExistException {
        editLDAPGroup(uri, name, description, serverURL, ldapGroupDefinition, field, userID, false, null, null);
    }

    /**
     * @ejb.interface-method
     * @ejb.transaction type="Required"
     */
    public void editLDAPGroup(URI uri, String name, String description, String serverURL, String ldapGroupDefinition, String field, String userID,
        boolean secured) throws MembershipServiceException, LibresourceSecurityException, URINotExistException {
        editLDAPGroup(uri, name, description, serverURL, ldapGroupDefinition, field, userID, secured, null, null);
    }

    /**
     * @ejb.interface-method
     * @ejb.transaction type="Required"
     */
    public void editLDAPGroup(URI uri, String name, String description, String serverURL, String ldapGroupDefinition, String field, String userID,
        boolean secured, String dnBind, String passBind)
        throws MembershipServiceException, LibresourceSecurityException, URINotExistException {
        try {
            KernelService kernelService = (KernelService) Libresource.getService(KernelConstants.SERVICE);
            LibresourceResourceIdentifier resourceIdentifier = kernelService.lookup(uri);

            if (!kernelService.checkSecurity(uri, KernelConstants.SECURITY_UPDATE)) {
                throw new LibresourceSecurityException(uri, KernelConstants.SECURITY_UPDATE);
            }

            LDAPGroupResourceLocal group = (LDAPGroupResourceLocal) Libresource.findResource(resourceIdentifier, LDAPGroupResourceLocal.class);
            group.setName(name);
            group.setDescription(description);
            group.setLDAPServerURL(serverURL);
            group.setLDAPGroupDefinition(ldapGroupDefinition);
            group.setField(field);
            group.setUserID(userID);

            group.setConnectionSecured(secured);
            group.setDNBind(dnBind);
            group.setPasswordBind(passBind);

            // update node
            kernelService.setShortName(uri, name);
            kernelService.setUpdateDate(uri, new Date());

            // event
            LibresourceEvent event = new LibresourceEvent(uri, resourceIdentifier, kernelService.getConnectedResource(), MembershipConstants.EVENT_GROUP_EDIT);
            Libresource.throwEvent(event);

            // indexation
            Libresource.index(uri);
        } catch (LibresourceSecurityException se) {
            ctx.setRollbackOnly();
            throw se;
        } catch (URINotExistException se) {
            ctx.setRollbackOnly();
            throw se;
        } catch (Exception e) {
            ctx.setRollbackOnly();
            throw new MembershipServiceException("Error in editGroup : " + e.getMessage(), e);
        }
    }

    /**
     * @ejb.interface-method
     * @ejb.transaction type="Required"
     */
    public void deleteGroup(URI uri) throws MembershipServiceException, LibresourceSecurityException, URINotExistException {
        try {
            KernelService kernelService = (KernelService) Libresource.getService(KernelConstants.SERVICE);
            LibresourceResourceIdentifier resourceIdentifier = kernelService.lookup(uri);

            Libresource.checkType(resourceIdentifier, GroupResourceLocal.class);
            kernelService.deleteURI(uri);

            // event
            LibresourceEvent event = new LibresourceEvent(uri, resourceIdentifier, kernelService.getConnectedResource(), MembershipConstants.EVENT_GROUP_EDIT);
            Libresource.throwEvent(event);
        } catch (LibresourceSecurityException se) {
            ctx.setRollbackOnly();
            throw se;
        } catch (URINotExistException se) {
            ctx.setRollbackOnly();
            throw se;
        } catch (Exception e) {
            ctx.setRollbackOnly();
            throw new MembershipServiceException("Error in deleteGroup : " + e.getMessage(), e);
        }
    }

    /**
     * @ejb.interface-method
     * @ejb.transaction type="Required"
     */
    public void deleteLDAPGroup(URI uri) throws MembershipServiceException, LibresourceSecurityException, URINotExistException {
        try {
            KernelService kernelService = (KernelService) Libresource.getService(KernelConstants.SERVICE);
            LibresourceResourceIdentifier resourceIdentifier = kernelService.lookup(uri);

            Libresource.checkType(resourceIdentifier, LDAPGroupResourceLocal.class);
            kernelService.deleteURI(uri);

            // event
            LibresourceEvent event = new LibresourceEvent(uri, resourceIdentifier, kernelService.getConnectedResource(), MembershipConstants.EVENT_GROUP_DELETE);
            Libresource.throwEvent(event);
        } catch (LibresourceSecurityException se) {
            ctx.setRollbackOnly();
            throw se;
        } catch (URINotExistException se) {
            ctx.setRollbackOnly();
            throw se;
        } catch (Exception e) {
            ctx.setRollbackOnly();
            throw new MembershipServiceException("Error in deleteGroup : " + e.getMessage(), e);
        }
    }

    /**
     * @ejb.interface-method
     * @ejb.transaction type="Supports"
     */
    public LibresourceResourceValue[] listMembers(URI groupUri)
        throws MembershipServiceException, LibresourceSecurityException, URINotExistException {
        try {
            KernelService kernelService = (KernelService) Libresource.getService(KernelConstants.SERVICE);
            LibresourceResourceIdentifier resourceIdentifier = kernelService.lookup(groupUri);
            Vector members = new Vector();
            URI memberUri = null;

            if (resourceIdentifier.getResourceType().equals(MembershipConstants.RESOURCE_GROUP)) {
                GroupResourceLocal group = (GroupResourceLocal) Libresource.findResource(resourceIdentifier, GroupResourceLocal.class);
                String[] memberIds = group.getMembers();

                for (int i = 0; i < memberIds.length; i++) {
                    String memberId = memberIds[i];

                    if (memberId.startsWith("User:")) {
                        memberUri = kernelService.normalizeURI(new URI(getUsersRootURI().getPath() + "/" + memberId.substring(5)));
                    } else if (memberId.startsWith("Group:")) {
                        try {
                            memberUri = new URI(memberId.substring(6));

                            // memberUri = kernelService.getMyURI(new
                            // LibresourceResourceIdentifier("Membership",
                            // "Group",
                            // memberId.substring(6)));
                        } catch (Exception e) {
                            //
                        }
                    }

                    try {
                        members.add(kernelService.getResource(memberUri));
                    } catch (Exception e) {
                        //
                    }
                }
            } else if (resourceIdentifier.getResourceType().equals(MembershipConstants.RESOURCE_LDAPGROUP)) {
                LDAPGroupResourceLocal group = (LDAPGroupResourceLocal) Libresource.findResource(resourceIdentifier, LDAPGroupResourceLocal.class);
                Collection logins = listLDAPMember(group);

                for (Iterator loginIterator = logins.iterator(); loginIterator.hasNext();) {
                    memberUri = kernelService.normalizeURI(new URI(getUsersRootURI().getPath() + "/" + loginIterator.next()));

                    try {
                        members.add(kernelService.getResource(memberUri));
                    } catch (Exception e) {
                    }
                }
            }

            LibresourceResourceValue[] result = new LibresourceResourceValue[members.size()];

            for (int i = 0; i < members.size(); i++) {
                result[i] = (LibresourceResourceValue) members.get(i);
            }

            return result;
        } catch (LibresourceSecurityException se) {
            throw se;
        } catch (URINotExistException se) {
            throw se;
        } catch (Exception e) {
            throw new MembershipServiceException("Error in listMembers : " + e.getMessage(), e);
        }
    }

    /**
     * @ejb.interface-method
     * @ejb.transaction type="Required"
     */
    public void addMemberInGroup(URI groupURI, URI memberUri)
        throws MembershipServiceException, LibresourceSecurityException, URINotExistException {
        try {
            KernelService kernelService = (KernelService) Libresource.getService(KernelConstants.SERVICE);
            LibresourceResourceIdentifier groupResourceIdentifier = kernelService.lookup(groupURI);

            if (!kernelService.checkSecurity(groupURI, KernelConstants.SECURITY_UPDATE)) {
                throw new LibresourceSecurityException(groupURI, KernelConstants.SECURITY_UPDATE);
            }

            LibresourceResourceIdentifier memberResourceIdentifier = null;

            try {
                memberResourceIdentifier = kernelService.systemLookup(memberUri);
            } catch (Exception e) {
                throw new URINotExistException(memberUri);
            }

            GroupResourceLocal groupResourceLocal = (GroupResourceLocal) Libresource.findResource(groupResourceIdentifier, GroupResourceLocal.class);

            if (memberResourceIdentifier == null) {
                throw new MembershipServiceException("Invalid URI");
            }

            if (!memberResourceIdentifier.getService().equals(MembershipConstants.SERVICE)) {
                throw new MembershipServiceException("Only User and Group can be member of a group");
            }

            String memberId = "";

            if (memberResourceIdentifier.getResourceType().equals(MembershipConstants.RESOURCE_PROFILE)) {
                memberId = "User:" + memberResourceIdentifier.getResourceId();
            } else {
                memberId = "Group:" + memberUri.getPath();
            }

            groupResourceLocal.addMember(memberId);

            // update node
            kernelService.setUpdateDate(groupURI, new Date());

            // event
            LibresourceEvent event = new LibresourceEvent(groupURI, groupResourceIdentifier, kernelService.getConnectedResource(),
                    MembershipConstants.EVENT_GROUP_ADDMEMBER, memberId);
            Libresource.throwEvent(event);
        } catch (LibresourceSecurityException se) {
            ctx.setRollbackOnly();
            throw se;
        } catch (URINotExistException se) {
            ctx.setRollbackOnly();
            throw se;
        } catch (Exception e) {
            ctx.setRollbackOnly();
            throw new MembershipServiceException("Error in addMemberInGroup : " + e.getMessage(), e);
        }
    }

    /**
     * @ejb.interface-method
     * @ejb.transaction type="Required"
     */
    public void removeMemberFromGroup(URI groupURI, URI memberUri)
        throws MembershipServiceException, LibresourceSecurityException, URINotExistException {
        try {
            KernelService kernelService = (KernelService) Libresource.getService(KernelConstants.SERVICE);
            LibresourceResourceIdentifier groupResourceIdentifier = kernelService.lookup(groupURI);
            LibresourceResourceIdentifier memberResourceIdentifier = kernelService.systemLookup(memberUri);

            if (!kernelService.checkSecurity(groupURI, KernelConstants.SECURITY_UPDATE)) {
                throw new LibresourceSecurityException(groupURI, KernelConstants.SECURITY_UPDATE);
            }

            GroupResourceLocal groupResourceLocal = (GroupResourceLocal) Libresource.findResource(groupResourceIdentifier, GroupResourceLocal.class);

            if (!memberResourceIdentifier.getService().equals(MembershipConstants.SERVICE)) {
                throw new MembershipServiceException("Only User and Group can be member of a group");
            }

            String memberId = "";

            if (memberResourceIdentifier.getResourceType().equals(MembershipConstants.RESOURCE_PROFILE)) {
                memberId = "User:" + memberResourceIdentifier.getResourceId();
            } else {
                memberId = "Group:" + memberUri.getPath();
            }

            groupResourceLocal.removeMember(memberId);

            // update node
            kernelService.setUpdateDate(groupURI, new Date());

            // event
            LibresourceEvent event = new LibresourceEvent(groupURI, groupResourceIdentifier, kernelService.getConnectedResource(),
                    MembershipConstants.EVENT_GROUP_REMOVEMEMBER, memberId);
            Libresource.throwEvent(event);
        } catch (LibresourceSecurityException se) {
            ctx.setRollbackOnly();
            throw se;
        } catch (URINotExistException se) {
            ctx.setRollbackOnly();
            throw se;
        } catch (Exception e) {
            ctx.setRollbackOnly();
            throw new MembershipServiceException("Error in removeMemberFromGroup : " + e.getMessage(), e);
        }
    }

    /**
     * @ejb.interface-method
     * @ejb.transaction type="Supports"
     */
    public boolean isUserInGroup(URI groupURI, URI userURI)
        throws MembershipServiceException, LibresourceSecurityException, URINotExistException {
        try {
            KernelService kernelService = (KernelService) Libresource.getService(KernelConstants.SERVICE);
            LibresourceResourceIdentifier groupResourceIdentifier = kernelService.lookup(groupURI);
            LibresourceResourceIdentifier userResourceIdentifier = kernelService.lookup(userURI);

            // GroupResourceLocal groupResourceLocal = (GroupResourceLocal)
            // Libresource.findResource(groupResourceIdentifier,
            // GroupResourceLocal.class);
            if (!userResourceIdentifier.getService().equals(MembershipConstants.SERVICE) ||
                    !userResourceIdentifier.getResourceType().equals(MembershipConstants.RESOURCE_PROFILE)) {
                throw new MembershipServiceException("No user at uri " + userURI);
            }

            String userId = userResourceIdentifier.getResourceId();

            return isMember(userId, groupURI, new Vector());
        } catch (LibresourceSecurityException se) {
            throw se;
        } catch (URINotExistException se) {
            throw se;
        } catch (Exception e) {
            throw new MembershipServiceException("Error in isUserInGroup : " + e.getMessage(), e);
        }
    }

    private boolean isMember(String userId, URI groupUri, Vector cycleDetection)
        throws Exception {
        try {
            KernelService kernelService = (KernelService) Libresource.getService(KernelConstants.SERVICE);
            GroupResourceLocal group = (GroupResourceLocal) Libresource.findResource(kernelService.systemLookup(kernelService.normalizeAbsoluteURIPath(groupUri)));

            cycleDetection.add(group.getId());

            if (groupUri.getPath().equals(getAllAuthentifiedUserGroupURI().getPath()) && !userId.equals(getUnauthentifiedUserId())) {
                return true;
            }

            if (group.isMember("User:" + userId) || group.isMember("User:" + getUnauthentifiedUserId())) {
                return true;
            }

            for (Iterator i = group.getGroups().iterator(); i.hasNext();) {
                URI grpUri = (URI) i.next();
                GroupResourceLocal groupResourceLocal = (GroupResourceLocal) Libresource.findResource(kernelService.systemLookup(kernelService.normalizeURI(
                                grpUri)));

                if (!cycleDetection.contains(groupResourceLocal.getId()) && isMember(userId, grpUri, cycleDetection)) {
                    return true;
                }
            }

            return false;
        } catch (URINotExistException e) {
            return false;
        }
    }

    /**
     * @ejb.interface-method
     * @ejb.transaction type="Supports"
     */
    public ProfileResourceValue getProfile(URI uri) throws MembershipServiceException, LibresourceSecurityException, URINotExistException {
        try {
            KernelService kernelService = (KernelService) Libresource.getService(KernelConstants.SERVICE);
            LibresourceResourceIdentifier resourceIdentifier = kernelService.lookup(uri);
            ProfileResourceLocal profileResourceLocal = (ProfileResourceLocal) Libresource.findResource(resourceIdentifier, ProfileResourceLocal.class);
            ProfileResourceValue resourceValue = profileResourceLocal.getProfileResourceValue();

            if ((resourceValue.getFullName() == null) || (resourceValue.getFullName().trim().length() == 0)) {
                resourceValue.setFullName(resourceValue.getId());
            }

            resourceValue.setUri(kernelService.normalizeURI(uri));

            return resourceValue;
        } catch (LibresourceSecurityException se) {
            throw se;
        } catch (URINotExistException se) {
            throw se;
        } catch (Exception e) {
            throw new MembershipServiceException("Error in getProfile : " + e.getMessage(), e);
        }
    }

    /**
     * @ejb.interface-method
     * @ejb.transaction type="Supports"
     */
    public ProfileResourceValue systemGetProfile(URI uri)
        throws MembershipServiceException, LibresourceSecurityException, URINotExistException {
        try {
            KernelService kernelService = (KernelService) Libresource.getService(KernelConstants.SERVICE);
            LibresourceResourceIdentifier resourceIdentifier = kernelService.systemLookup(uri);
            ProfileResourceLocal profileResourceLocal = (ProfileResourceLocal) Libresource.findResource(resourceIdentifier, ProfileResourceLocal.class);
            ProfileResourceValue resourceValue = profileResourceLocal.getProfileResourceValue();
            resourceValue.setUri(kernelService.normalizeURI(uri));

            return resourceValue;
        } catch (LibresourceSecurityException se) {
            throw se;
        } catch (URINotExistException se) {
            throw se;
        } catch (Exception e) {
            throw new MembershipServiceException("Error in systemGetProfile : " + e.getMessage(), e);
        }
    }

    /**
     * @ejb.interface-method
     * @ejb.transaction type="Supports"
     */
    public String systemGetUserPassword(URI uri) throws LibresourceException, URINotExistException {
        Connection conn = null;
        Statement stmt = null;

        try {
            getConnectionForUsers();
            conn = Libresource.getDatasource().getConnection();
            stmt = conn.createStatement();

            ResultSet rs = stmt.executeQuery("SELECT passworddigest_ FROM casusers_ WHERE username_='" + UriManager.getPathName(uri) + "';");

            if (!rs.next()) {
                throw new URINotExistException("Username does not exists !");
            }

            return rs.getString("passworddigest_");
        } catch (Exception e) {
            throw new LibresourceException(e.getMessage());
        } finally {
            try {
                if (stmt != null) {
                    stmt.close();
                }
            } catch (Exception e) {
            }

            try {
                if (conn != null) {
                    conn.close();
                }
            } catch (Exception e) {
            }
        }
    }

    /**
     * @ejb.interface-method
     * @ejb.transaction type="Supports"
     */
    public ProfileResourceValue getProfile() throws MembershipServiceException, LibresourceSecurityException, URINotExistException {
        try {
            KernelService kernelService = (KernelService) Libresource.getService(KernelConstants.SERVICE);

            return systemGetProfile(kernelService.getConnectedResource());
        } catch (LibresourceSecurityException se) {
            throw se;
        } catch (URINotExistException se) {
            throw se;
        } catch (Exception e) {
            throw new MembershipServiceException("Error in getProfile : " + e.getMessage(), e);
        }
    }

    /**
     * @ejb.interface-method
     * @ejb.transaction type="Supports"
     */
    public ProfileResourceValue getProfile(String userId)
        throws MembershipServiceException, LibresourceSecurityException, URINotExistException {
        try {
            return getProfile(new URI(getUsersRootURI() + "/" + userId));
        } catch (LibresourceSecurityException se) {
            throw se;
        } catch (URINotExistException se) {
            throw se;
        } catch (Exception e) {
            throw new MembershipServiceException("Error in getProfile : " + e.getMessage(), e);
        }
    }

    /**
     * @ejb.interface-method
     * @ejb.transaction type="Supports"
     */
    public Vector listUsers() throws MembershipServiceException, LibresourceSecurityException, URINotExistException {
        try {
            KernelService kernelService = (KernelService) Libresource.getService(KernelConstants.SERVICE);
            URI[] usersUris = kernelService.listChildren(getUsersRootURI());
            Vector users = new Vector();

            for (int i = 0; i < usersUris.length; i++) {
                try {
                    ProfileResourceValue profile = getProfile(usersUris[i]);
                    users.add(profile);
                } catch (Exception e) {
                    //
                }
            }

            return users;
        } catch (LibresourceSecurityException se) {
            throw se;
        } catch (URINotExistException se) {
            throw se;
        } catch (Exception e) {
            throw new MembershipServiceException("Error in listUsers : " + e.getMessage(), e);
        }
    }

    // TODO check in mailing list
    // /**
    // * @ejb.interface-method
    // * @ejb.transaction type="Supports"
    // */
    // public String getUserName(String userMail) throws
    // MembershipServiceException, LibresourceSecurityException,
    // URINotExistException {
    // try {
    // return
    // InternalUserBaseUtil.getLocalHome().create().getUserName(userMail);
    // } catch (LibresourceSecurityException se) {
    // throw se;
    // } catch (URINotExistException se) {
    // throw se;
    // } catch (Exception e) {
    // throw new MembershipServiceException("Error in getUserName : " +
    // e.getMessage(), e);
    // }
    // }

    /**
     * @ejb.interface-method
     * @ejb.transaction type="Supports"
     */
    public GroupResourceValue systemGetGroup(URI uri) throws MembershipServiceException, LibresourceSecurityException, URINotExistException {
        try {
            KernelService kernelService = (KernelService) Libresource.getService(KernelConstants.SERVICE);
            LibresourceResourceIdentifier resourceIdentifier = kernelService.systemLookup(uri);
            GroupResourceLocal groupResourceLocal = (GroupResourceLocal) Libresource.findResource(resourceIdentifier, GroupResourceLocal.class);
            GroupResourceValue resourceValue = groupResourceLocal.getGroupResourceValue();
            resourceValue.setUri(kernelService.normalizeURI(uri));

            return resourceValue;
        } catch (LibresourceSecurityException se) {
            throw se;
        } catch (URINotExistException se) {
            throw se;
        } catch (Exception e) {
            throw new MembershipServiceException("Error in getGroup : " + e.getMessage(), e);
        }
    }

    /**
     * @ejb.interface-method
     * @ejb.transaction type="Required"
     */
    public void systemCreateProfile(String userId, String userName, String userEmail, String userJabberId, HashMap userInfos)
        throws MembershipServiceException, LibresourceSecurityException {
        try {
            if (!userId.matches("[A-Za-z0-9]+")) {
                throw new Exception("invalid user id (userId should match [A-Za-z0-9])");
            }

            ProfileResourceLocal profileResourceLocal = ProfileResourceUtil.getLocalHome().create(userId, userName, userEmail, userJabberId, userInfos);
            KernelService kernelService = (KernelService) Libresource.getService(KernelConstants.SERVICE);

            URI profileUri = new URI(getUsersRootURI().getPath() + "/" + userId);
            kernelService.systemBind(profileResourceLocal.getLibresourceResourceIdentifier(), profileUri, userName);
            kernelService.systemCopyAcl(getUsersRootURI(), profileUri);
            kernelService.systemCreateAcl(profileUri, profileUri, new String[] { KernelConstants.SECURITY_READ, KernelConstants.SECURITY_UPDATE });

            // event
            LibresourceEvent event = new LibresourceEvent(profileUri, profileResourceLocal.getLibresourceResourceIdentifier(),
                    kernelService.getConnectedResource(), MembershipConstants.EVENT_PROFILE_CREATE);
            Libresource.throwEvent(event);
        } catch (LibresourceSecurityException se) {
            ctx.setRollbackOnly();
            throw se;
        } catch (Exception e) {
            ctx.setRollbackOnly();
            throw new MembershipServiceException("Error in systemCreateProfile : " + e.getMessage(), e);
        }
    }

    /**
     * @ejb.interface-method
     * @ejb.transaction type="Supports"
     */
    public GroupResourceValue getGroup(URI uri) throws MembershipServiceException, LibresourceSecurityException, URINotExistException {
        try {
            KernelService kernelService = (KernelService) Libresource.getService(KernelConstants.SERVICE);
            LibresourceResourceIdentifier resourceIdentifier = kernelService.lookup(uri);
            GroupResourceLocal groupResourceLocal = (GroupResourceLocal) Libresource.findResource(resourceIdentifier, GroupResourceLocal.class);
            GroupResourceValue resourceValue = groupResourceLocal.getGroupResourceValue();
            resourceValue.setUri(kernelService.normalizeURI(uri));

            return resourceValue;
        } catch (LibresourceSecurityException se) {
            throw se;
        } catch (URINotExistException se) {
            throw se;
        } catch (Exception e) {
            throw new MembershipServiceException("Error in getGroup : " + e.getMessage(), e);
        }
    }

    /**
     * @ejb.interface-method
     * @ejb.transaction type="Supports"
     */
    public LDAPGroupResourceValue getLDAPGroup(URI uri)
        throws MembershipServiceException, LibresourceSecurityException, URINotExistException {
        try {
            KernelService kernelService = (KernelService) Libresource.getService(KernelConstants.SERVICE);
            LibresourceResourceIdentifier resourceIdentifier = kernelService.lookup(uri);
            LDAPGroupResourceLocal groupResourceLocal = (LDAPGroupResourceLocal) Libresource.findResource(resourceIdentifier, LDAPGroupResourceLocal.class);
            LDAPGroupResourceValue resourceValue = groupResourceLocal.getLDAPGroupResourceValue();
            resourceValue.setUri(kernelService.normalizeURI(uri));

            return resourceValue;
        } catch (LibresourceSecurityException se) {
            throw se;
        } catch (URINotExistException se) {
            throw se;
        } catch (Exception e) {
            throw new MembershipServiceException("Error in getLDAPGroup : " + e.getMessage(), e);
        }
    }

    // libresource service

    /**
     * @ejb.interface-method
     * @ejb.transaction type="Supports"
     */
    public boolean canMapResourceForSecurity(URI connectedUri, URI toUri)
        throws LibresourceException {
        try {
            KernelService kernelService = (KernelService) Libresource.getService(KernelConstants.SERVICE);
            LibresourceResourceIdentifier connectedResourceIdentifier = kernelService.systemLookup(connectedUri);
            LibresourceResourceIdentifier toResourceIdentifier = kernelService.systemLookup(toUri);

            if (connectedResourceIdentifier.getService().equals(MembershipConstants.SERVICE) &&
                    connectedResourceIdentifier.getResourceType().equals(MembershipConstants.RESOURCE_PROFILE)) {
                String userId = connectedResourceIdentifier.getResourceId();

                if (toResourceIdentifier == null) {
                    return false;
                }

                if (toResourceIdentifier.getService().equals(MembershipConstants.SERVICE) &&
                        toResourceIdentifier.getResourceType().equals(MembershipConstants.RESOURCE_PROFILE)) {
                    return connectedResourceIdentifier.equals(toResourceIdentifier) || toResourceIdentifier.getResourceId().equals(getUnauthentifiedUserId());
                } else if (toResourceIdentifier.getService().equals(MembershipConstants.SERVICE) &&
                        toResourceIdentifier.getResourceType().equals(MembershipConstants.RESOURCE_GROUP)) {
                    // GroupResourceLocal groupResourceLocal =
                    // (GroupResourceLocal)
                    // Libresource.findResource(toResourceIdentifier);
                    return isMember(userId, toUri, new Vector());
                } else if (toResourceIdentifier.getService().equals(MembershipConstants.SERVICE) &&
                        toResourceIdentifier.getResourceType().equals(MembershipConstants.RESOURCE_LDAPGROUP)) {
                    // Check if user in LDAP group
                    // FIXME LDAP
                    LDAPGroupResourceLocal ldapGroup = (LDAPGroupResourceLocal) Libresource.findResource(toResourceIdentifier);
                    System.out.println("==> userid= " + userId);

                    return isLDAPMember(userId, ldapGroup);
                }
            }

            if (connectedResourceIdentifier.getService().equals(MembershipConstants.SERVICE) &&
                    connectedResourceIdentifier.getResourceType().equals(MembershipConstants.RESOURCE_GROUP)) {
                if (toResourceIdentifier == null) {
                    return false;
                }

                if (toResourceIdentifier.getService().equals(MembershipConstants.SERVICE) &&
                        toResourceIdentifier.getResourceType().equals(MembershipConstants.RESOURCE_GROUP)) {
                    return connectedResourceIdentifier.equals(toResourceIdentifier);
                }
            }

            return false;
        } catch (Exception e) {
            throw new LibresourceException("error in MembershipService.canMapConnectedResource(" + connectedUri + "," + toUri + ")", e);
        }
    }

    /**
     * @ejb.interface-method
     * @ejb.transaction type="Supports"
     */
    public LibresourceIndexableContent getIndexableContent(LibresourceResourceIdentifier resourceIdentifier)
        throws LibresourceException {
        try {
            LibresourceIndexableContent content = new LibresourceIndexableContent();

            try {
                Libresource.checkType(resourceIdentifier, ProfileResourceLocal.class);

                ProfileResourceLocal profile = (ProfileResourceLocal) Libresource.findResource(resourceIdentifier);
                content.addContentPart(profile.getId());
                content.addContentPart(profile.getFullName());
                content.addContentPart(profile.getEmail());
                content.addContentPart(profile.getJabberId());
            } catch (LibresourceException e) {
                //
            }

            try {
                Libresource.checkType(resourceIdentifier, GroupResourceLocal.class);

                GroupResourceLocal groupResourceLocal = (GroupResourceLocal) Libresource.findResource(resourceIdentifier);
                content.addContentPart(groupResourceLocal.getName());
                content.addContentPart(groupResourceLocal.getDescription());
            } catch (LibresourceException e) {
                //
            }

            return content;
        } catch (Exception e) {
            throw new LibresourceException("error in Membership.getIndexableContent(" + resourceIdentifier + ")", e);
        }
    }

    /**
     * @ejb.interface-method
     * @ejb.transaction type="Supports"
     */
    public LibresourceExportHandler getXmlExportHandlerForResource(URI uri)
        throws LibresourceException {
        try {
            KernelService kernelService = (KernelService) Libresource.getService(KernelConstants.SERVICE);
            LibresourceResourceIdentifier resourceIdentifier = kernelService.lookup(uri);

            try {
                Libresource.checkType(resourceIdentifier, GroupResourceLocal.class);

                GroupExportHandler exportHandler = new GroupExportHandler(uri);

                return exportHandler;
            } catch (Exception e) {
                //
            }

            try {
                Libresource.checkType(resourceIdentifier, ProfileResourceLocal.class);

                ProfileExportHandler exportHandler = new ProfileExportHandler(uri);

                return exportHandler;
            } catch (Exception e) {
                //
            }

            return super.getXmlExportHandlerForResource(uri);
        } catch (Exception e) {
            throw new LibresourceException("Can't obtain exportHandler for uri " + uri, e);
        }
    }

    /**
     * @ejb.interface-method
     * @ejb.transaction type="Supports"
     */
    public LibresourceImportHandler getXmlImportHandler(String type)
        throws LibresourceException {
        if (type.equals(MembershipConstants.RESOURCE_GROUP)) {
            return new GroupImportHandler();
        }

        if (type.equals(MembershipConstants.RESOURCE_PROFILE)) {
            return new ProfileImportHandler();
        }

        return super.getXmlImportHandler(type);
    }

    /**
     * @ejb.interface-method
     * @ejb.transaction type="Supports"
     */
    public String[] listAvailablesPermissions(LibresourceResourceIdentifier resourceIdentifier)
        throws LibresourceException {
        if (resourceIdentifier.getService().equals(MembershipConstants.SERVICE) &&
                resourceIdentifier.getResourceType().equals(MembershipConstants.RESOURCE_PROFILE)) {
            return null;
        }

        return super.listAvailablesPermissions(resourceIdentifier);
    }

    /**
     * @ejb.interface-method
     * @ejb.transaction type="Supports"
     */
    public String[] listAvailablesEvents(LibresourceResourceIdentifier resourceIdentifier)
        throws LibresourceException {
        if (resourceIdentifier.getService().equals(MembershipConstants.SERVICE)) {
            if (resourceIdentifier.getResourceType().equals(MembershipConstants.RESOURCE_GROUP)) {
                return new String[] {
                    MembershipConstants.EVENT_GROUP_CREATE, MembershipConstants.EVENT_GROUP_EDIT, MembershipConstants.EVENT_GROUP_ADDMEMBER,
                    MembershipConstants.EVENT_GROUP_REMOVEMEMBER
                };
            }

            if (resourceIdentifier.getResourceType().equals(MembershipConstants.RESOURCE_PROFILE)) {
                return new String[] { MembershipConstants.EVENT_PROFILE_CREATE, MembershipConstants.EVENT_PROFILE_EDIT };
            }
        }

        return super.listAvailablesPermissions(resourceIdentifier);
    }

    /**
     * @ejb.interface-method
     * @ejb.transaction type="Required"
     */
    public void createCasUser(String username, String password)
        throws LibresourceException {
        Connection conn = null;
        Statement stmt = null;
        PreparedStatement insertUser = null;

        try {
            getConnectionForUsers();
            conn = Libresource.getDatasource().getConnection();
            stmt = conn.createStatement();

            ResultSet rs = stmt.executeQuery("SELECT * FROM casusers_ WHERE username_='" + username + "';");

            if (rs.next()) {
                throw new LibresourceException("Username already exists !");
            }

            if (!username.matches("[A-Za-z0-9]+")) {
                throw new LibresourceException("invalid user id (userId should match [A-Za-z0-9])");
            }

            // insert user
            insertUser = conn.prepareStatement("INSERT INTO casusers_ (username_, passworddigest_) VALUES (?,?);");
            insertUser.setString(1, username);

            // No more digest
            insertUser.setString(2, password); // digest(password));
            insertUser.execute();
            insertUser.close();

            // create profile
            KernelService kernelService = (KernelService) Libresource.getService(KernelConstants.SERVICE);
            kernelService.systemCreateURI(new URI(getUsersRootURI() + "/" + username));

            String jabberHost = Libresource.getLibresourceConfiguration("kernel.serverUrl");
            jabberHost = jabberHost.split(":")[1]; // "http://localhost:9000" => get "//localhost"
            jabberHost = jabberHost.substring(2);

            systemCreateProfile(username, username, "", username + "@" + jabberHost, new HashMap());
        } catch (Exception e) {
            throw new LibresourceException(e.getMessage());
        } finally {
            try {
                if (stmt != null) {
                    stmt.close();
                }
            } catch (Exception e) {
            }

            try {
                if (insertUser != null) {
                    insertUser.close();
                }
            } catch (Exception e) {
            }

            try {
                if (conn != null) {
                    conn.close();
                }
            } catch (Exception e) {
            }
        }
    }

    /**
     * @ejb.interface-method
     * @ejb.transaction type="Required"
     */
    public void editCasUser(String username, String password)
        throws LibresourceException {
	editCasUserLDAP(username,password);
    }

    public void editCasUserDB(String username, String password)
        throws LibresourceException {
        Connection conn = null;
        Statement stmt = null;
        PreparedStatement insertUser = null;

        try {
            KernelService kernelService = (KernelService) Libresource.getService(KernelConstants.SERVICE);

            if (!kernelService.checkSecurity(new URI(getUsersRootURI() + "/" + username), KernelConstants.SECURITY_UPDATE)) {
                throw new LibresourceException("Not allowed to change password");
            }

            getConnectionForUsers();
            conn = Libresource.getDatasource().getConnection();
            stmt = conn.createStatement();

            ResultSet rs = stmt.executeQuery("SELECT * FROM casusers_ WHERE username_='" + username + "';");

            if (!rs.next()) {
                throw new LibresourceException("Username \"" + username + "\" does not exists !");
            }

            // insert user
            insertUser = conn.prepareStatement("UPDATE casusers_ SET passworddigest_ = ? WHERE  username_= ?;");
            insertUser.setString(2, username);

            // No more need of digest
            insertUser.setString(1, password); //digest(password));
            insertUser.executeUpdate();
            insertUser.close();
        } catch (Exception e) {
            throw new LibresourceException(e.getMessage());
        } finally {
            try {
                if (stmt != null) {
                    stmt.close();
                }
            } catch (Exception e) {
            }

            try {
                if (insertUser != null) {
                    insertUser.close();
                }
            } catch (Exception e) {
            }

            try {
                if (conn != null) {
                    conn.close();
                }
            } catch (Exception e) {
            }
        }
    }

    public void editCasUserLDAP(String username, String password)
        throws LibresourceException {
        LdapContext searchCtx = null;

        try {
            KernelService kernelService = (KernelService) Libresource.getService(KernelConstants.SERVICE);

            if (!kernelService.checkSecurity(new URI(getUsersRootURI() + "/" + username), KernelConstants.SECURITY_UPDATE)) {
                throw new LibresourceException("Not allowed to change password");
            }

            searchCtx = getLDAPConnection();
	    String searchBase = Libresource.getLibresourceConfiguration("ldap.search-base");
	    String user = "uid=" + username;
	    final String[] returnedAttributes = { 
		"cn" 
	    };
		
	    // create a new searchControl object with the parameters we have set
	    SearchControls constraints = new SearchControls(2,
							    // Maximum number to return, 0 -> no limit
							    1,
							    // Number of ms to wait before return, 0 -> infinite
							    1000,
							    // attributes to return
							    returnedAttributes,
							    // return the object bound to the name
							    false,
							    // deference the link during search
							    false);

	    // search in the LDAP directory
	    constraints.setSearchScope(SearchControls.SUBTREE_SCOPE);
	    NamingEnumeration result = searchCtx.search(searchBase,user,constraints);
	    SearchResult searchResult = (SearchResult) result.next();
            if (searchResult == null) {
                throw new LibresourceException("Username \"" + username + "\" does not exists !");
            }

	    String url = Libresource.getLibresourceConfiguration("ldap.url");
	    String userDn = searchResult.getName() + "," + searchBase;

	    PasswordChangeResponse pcr = (PasswordChangeResponse) searchCtx.extendedOperation(new PasswordChangeRequest(userDn, null, password));


        } catch (Exception e) {
            throw new LibresourceException(e.getMessage());
        } finally {
            try {
                if (searchCtx != null) {
                    searchCtx.close();
                }
            } catch (Exception e) {
            }
        }
    }

    private void getConnectionForUsers() throws Exception {
        Connection conn = null;
        Statement stmt = null;
        ResultSet rs = null;

        try {
            conn = Libresource.getDatasource().getConnection();
            stmt = conn.createStatement();
            rs = stmt.executeQuery("SELECT tablename FROM pg_tables WHERE tablename='casusers_';");

            // if table does not exist
            if (!rs.next()) {
                // create table
                stmt = conn.createStatement();
                stmt.execute("CREATE TABLE casusers_ (username_ VARCHAR PRIMARY KEY, passworddigest_ VARCHAR);");
            }
        } catch (Exception e) {
            throw e;
        } finally {
            try {
                if (stmt != null) {
                    stmt.close();
                }
            } catch (Exception e) {
            }

            try {
                if (rs != null) {
                    rs.close();
                }
            } catch (Exception e) {
            }

            try {
                if (conn != null) {
                    conn.close();
                }
            } catch (Exception e) {
            }
        }
    }

    public static String digest(String password) throws Exception {
        MessageDigest digest = MessageDigest.getInstance("MD5");
        digest.update(password.getBytes("UTF-8"));

        byte[] md5 = digest.digest();
        String sReturnMsg = "";

        for (int i = 0; i < md5.length; i++) {
            sReturnMsg += Integer.toHexString(md5[i]);
        }

        return sReturnMsg;
    }

    private void getConnectionForSubversionUsers() throws Exception {
        Connection conn = null;
        Statement stmt = null;
        ResultSet rs = null;

        try {
            conn = Libresource.getDatasource().getConnection();
            stmt = conn.createStatement();
            rs = stmt.executeQuery("SELECT tablename FROM pg_tables WHERE tablename='subversionusers_';");

            // if table does not exist
            if (!rs.next()) {
                // create table
                stmt = conn.createStatement();
                stmt.execute("CREATE TABLE subversionusers_ (username_ VARCHAR PRIMARY KEY, password_ VARCHAR);");
            }
        } catch (Exception e) {
            throw e;
        } finally {
            try {
                if (stmt != null) {
                    stmt.close();
                }
            } catch (Exception e) {
            }

            try {
                if (rs != null) {
                    rs.close();
                }
            } catch (Exception e) {
            }

            try {
                if (conn != null) {
                    conn.close();
                }
            } catch (Exception e) {
            }
        }
    }

    private void registerSubversionUser(String login, String password)
        throws Exception {
        Connection conn = null;
        PreparedStatement insertUser = null;
        ResultSet rs = null;

        try {
            getConnectionForSubversionUsers();
            insertUser = null;
            conn = Libresource.getDatasource().getConnection();

            // insert user
            insertUser = conn.prepareStatement("SELECT * FROM subversionusers_ WHERE username_ = ?");
            insertUser.setString(1, login);
            rs = insertUser.executeQuery();
            insertUser.close();

            if (!rs.next()) {
                insertUser = conn.prepareStatement("INSERT INTO subversionusers_ (username_, password_) VALUES (?,?);");
                insertUser.setString(1, login);
                insertUser.setString(2, password);
                insertUser.executeUpdate();
            } else {
                insertUser = conn.prepareStatement("UPDATE subversionusers_ SET password_ = ? WHERE username_ = ? ;");
                insertUser.setString(2, login);
                insertUser.setString(1, password);
                insertUser.executeUpdate();
            }
        } catch (Exception e) {
            throw e;
        } finally {
            try {
                if (insertUser != null) {
                    insertUser.close();
                }
            } catch (Exception e) {
            }

            try {
                if (rs != null) {
                    rs.close();
                }
            } catch (Exception e) {
            }

            try {
                if (conn != null) {
                    conn.close();
                }
            } catch (Exception e) {
            }
        }
    }

    private DirContext getLDAPConnection(LDAPGroupResourceLocal ldapGroup)
        throws Exception {
        DirContext connection = null;

        // initialisation
        String url = ldapGroup.getLDAPServerURL();
        String bindDn = ldapGroup.getDNBind();
        String bindPassword = ldapGroup.getPasswordBind();
        boolean secured = ldapGroup.getConnectionSecured();

        try {
            Hashtable hashtable = new Hashtable(5, 0.75f);
            hashtable.put(Context.INITIAL_CONTEXT_FACTORY, "com.sun.jndi.ldap.LdapCtxFactory");
            hashtable.put(Context.PROVIDER_URL, url);
            hashtable.put(Context.SECURITY_AUTHENTICATION, "simple");

            if (bindDn != null) {
                hashtable.put(Context.SECURITY_PRINCIPAL, bindDn);
                hashtable.put(Context.SECURITY_CREDENTIALS, bindPassword.getBytes());
            }

            if (secured) {
                hashtable.put(Context.SECURITY_PROTOCOL, "ssl");
            }

            System.out.println("Connecting to the LDAP directory (url=`" + url + "', username=`" + bindDn + "')...");
            connection = new InitialDirContext(hashtable);
            System.out.println("Connection succeeded.");

            return connection;
        } catch (AuthenticationException e) {
            System.out.println("Connection failed: " + e.getMessage());
        } catch (NamingException e) {
            System.out.println("Could not connect to \"" + url + "\"");
            System.out.println("Connection failed: '" + e.getMessage() + "')");
        }

        // connection failed, but try to close the connection however
        if (connection != null) {
            try {
                System.out.println("Closing LDAP connection...");
                connection.close();
            } catch (NamingException e2) {
                System.out.println("Could not close LDAP connection.");
            }
        }

        return null;
    }

    private LdapContext getLDAPConnection()
        throws Exception {
        LdapContext connection = null;

        // initialisation
	String url = Libresource.getLibresourceConfiguration("ldap.url");
        String bindDn = Libresource.getLibresourceConfiguration("ldap.bind-dn");
        String bindPassword = Libresource.getLibresourceConfiguration("ldap.bind-password");

        try {
            Hashtable hashtable = new Hashtable();
            hashtable.put(Context.INITIAL_CONTEXT_FACTORY, "com.sun.jndi.ldap.LdapCtxFactory");
            hashtable.put(Context.PROVIDER_URL, url);
            hashtable.put(Context.SECURITY_AUTHENTICATION, "simple");

            if (bindDn != null) {
                hashtable.put(Context.SECURITY_PRINCIPAL, bindDn);
                hashtable.put(Context.SECURITY_CREDENTIALS, bindPassword.getBytes());
            }
	    boolean secured = url.substring(0, 5).equals("ldaps");
	    if (secured) {
		hashtable.put(Context.SECURITY_PROTOCOL, "ssl");
	    }

            System.out.println("Connecting to the LDAP directory (url=`" + url + "', username=`" + bindDn + "')...");
            connection = new InitialLdapContext(hashtable,null);
            System.out.println("Connection succeeded.");

            return connection;
        } catch (AuthenticationException e) {
            System.out.println("Connection failed: " + e.getMessage());
        } catch (NamingException e) {
            System.out.println("Could not connect to \"" + url + "\"");
            System.out.println("Connection failed: '" + e.getMessage() + "')");
        }

        // connection failed, but try to close the connection however
        if (connection != null) {
            try {
                System.out.println("Closing LDAP connection...");
                connection.close();
            } catch (NamingException e2) {
                System.out.println("Could not close LDAP connection.");
            }
        }

        return null;
    }

    private boolean isLDAPMember(String userLogin, LDAPGroupResourceLocal ldapGroup)
        throws Exception {
        DirContext ldapConnection = getLDAPConnection(ldapGroup);

        if (ldapConnection != null) {
            Attributes attrs = ldapConnection.getAttributes(ldapGroup.getLDAPGroupDefinition());
            Attribute attr = attrs.get(ldapGroup.getField());
            NamingEnumeration members = attr.getAll();
            String currentUser = null;

            while (members.hasMore()) {
                currentUser = (String) members.next();

                if ((currentUser.indexOf(ldapGroup.getUserID()) != -1) && (currentUser.indexOf(userLogin) != -1)) {
                    // check if its real and not the same starting letter of the
                    // user login
                    currentUser = currentUser.substring(currentUser.indexOf(userLogin));

                    if (currentUser.split(",")[0].trim().equals(userLogin)) {
                        try {
                            ldapConnection.close();
                        } catch (Exception e) {
                        }

                        return true;
                    }
                }
            }

            try {
                ldapConnection.close();
            } catch (Exception e) {
            }
        }

        return false;
    }

    private Collection listLDAPMember(LDAPGroupResourceLocal ldapGroup)
        throws Exception {
        ArrayList result = new ArrayList();
        DirContext ldapConnection = getLDAPConnection(ldapGroup);

        if (ldapConnection != null) {
            Attributes attrs = ldapConnection.getAttributes(ldapGroup.getLDAPGroupDefinition());
            Attribute attr = attrs.get(ldapGroup.getField());
            NamingEnumeration members = attr.getAll();
            String currentUser = null;

            while (members.hasMore()) {
                currentUser = (String) members.next();

                if (currentUser.indexOf(ldapGroup.getUserID()) != -1) {
                    currentUser = currentUser.substring(currentUser.indexOf(ldapGroup.getUserID()));
                    currentUser = currentUser.substring(currentUser.indexOf("=") + 1);
                    result.add(currentUser.split(",")[0].trim());
                }
            }

            try {
                ldapConnection.close();
            } catch (Exception e) {
            }
        }

        return result;
    }
}
